from __future__ import annotations

import subprocess
from unittest.mock import Mock, call, patch

import pytest

from streamlink import Streamlink
from streamlink.stream.ffmpegmux import FFMPEGMuxer, FFmpegVersionOutput


# noinspection PyProtectedMember
@pytest.fixture(autouse=True)
def _resolve_command_cache_clear():
    FFMPEGMuxer._resolve_command.cache_clear()
    yield
    FFMPEGMuxer._resolve_command.cache_clear()


@pytest.fixture(autouse=True)
def _logger(caplog: pytest.LogCaptureFixture):
    caplog.set_level(1, "streamlink")


@pytest.fixture()
def session(session: Streamlink):
    session.set_option("ffmpeg-no-validation", True)

    return session


class TestCommand:
    def test_cache(self, session: Streamlink):
        with patch("streamlink.stream.ffmpegmux.which", return_value="some_value") as mock:
            assert FFMPEGMuxer.command(session) == "some_value"
            assert FFMPEGMuxer.command(session) == "some_value"
            assert len(mock.call_args_list) == 1
        with patch("streamlink.stream.ffmpegmux.which", return_value="other_value") as mock:
            assert FFMPEGMuxer.command(session) == "some_value"
            assert len(mock.call_args_list) == 0

    @pytest.mark.parametrize(
        ("command", "which", "expected"),
        [
            pytest.param(None, {"ffmpeg": None}, None, id="resolver-negative"),
            pytest.param(None, {"ffmpeg": "ffmpeg"}, "ffmpeg", id="resolver-posix"),
            pytest.param(None, {"ffmpeg": "ffmpeg.exe"}, "ffmpeg.exe", id="resolver-windows"),
            pytest.param("custom", {"ffmpeg": "ffmpeg"}, None, id="custom-negative"),
            pytest.param("custom", {"ffmpeg": "ffmpeg", "custom": "custom"}, "custom", id="custom-positive"),
        ],
    )
    def test_no_cache(self, session: Streamlink, command: str | None, which: dict, expected: str | None):
        session.options.update({"ffmpeg-ffmpeg": command})
        with patch("streamlink.stream.ffmpegmux.which", side_effect=which.get):
            assert FFMPEGMuxer.command(session) == expected

    @pytest.mark.parametrize(
        ("resolved", "expected"),
        [
            pytest.param(None, False, id="negative"),
            pytest.param("ffmpeg", True, id="positive"),
        ],
    )
    def test_is_usable(self, session: Streamlink, resolved: str | None, expected: bool):
        with patch("streamlink.stream.ffmpegmux.which", return_value=resolved):
            assert FFMPEGMuxer.is_usable(session) is expected

    def test_log(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        with patch("streamlink.stream.ffmpegmux.which", return_value=None):
            assert not FFMPEGMuxer.is_usable(session)
            assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
                (
                    "ffmpegmux",
                    "warning",
                    "No valid FFmpeg binary was found. See the --ffmpeg-ffmpeg option.",
                ),
                (
                    "ffmpegmux",
                    "warning",
                    "Muxing streams is unsupported! Only a subset of the available streams can be returned!",
                ),
            ]
            assert not FFMPEGMuxer.is_usable(session)
            assert len(caplog.records) == 2

    def test_no_log(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        with patch("streamlink.stream.ffmpegmux.which", return_value="foo"):
            assert FFMPEGMuxer.is_usable(session)
            assert not caplog.records

    def test_validate_success(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        session.options.update({"ffmpeg-no-validation": False})

        class MyFFmpegVersionOutput(FFmpegVersionOutput):
            def run(self):
                self.onstdout(0, "ffmpeg version 0.0.0 suffix")
                self.onstdout(1, "foo")
                self.onstdout(2, "bar")
                return True

        with (
            patch("streamlink.stream.ffmpegmux.which", return_value="/usr/bin/ffmpeg"),
            patch("streamlink.stream.ffmpegmux.FFmpegVersionOutput", side_effect=MyFFmpegVersionOutput) as mock_versionoutput,
        ):
            result = FFMPEGMuxer.command(session)

        assert result == "/usr/bin/ffmpeg"
        assert mock_versionoutput.call_args_list == [
            call(["/usr/bin/ffmpeg", "-version"], timeout=4.0),
        ]
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("ffmpegmux", "trace", "Querying FFmpeg version: ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "debug", "ffmpeg version 0.0.0 suffix"),
            ("ffmpegmux", "debug", " foo"),
            ("ffmpegmux", "debug", " bar"),
        ]

    def test_validate_failure(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        session.options.update({"ffmpeg-no-validation": False})

        class MyFFmpegVersionOutput(FFmpegVersionOutput):
            def run(self):
                return False

        with (
            patch("streamlink.stream.ffmpegmux.which", return_value="/usr/bin/ffmpeg"),
            patch("streamlink.stream.ffmpegmux.FFmpegVersionOutput", side_effect=MyFFmpegVersionOutput) as mock_versionoutput,
        ):
            result = FFMPEGMuxer.command(session)

        assert result is None
        assert mock_versionoutput.call_args_list == [
            call(["/usr/bin/ffmpeg", "-version"], timeout=4.0),
        ]
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("ffmpegmux", "trace", "Querying FFmpeg version: ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "error", "Could not validate FFmpeg!"),
            ("ffmpegmux", "error", "Unexpected FFmpeg version output while running ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "warning", "No valid FFmpeg binary was found. See the --ffmpeg-ffmpeg option."),
            ("ffmpegmux", "warning", "Muxing streams is unsupported! Only a subset of the available streams can be returned!"),
        ]


class TestFFmpegVersionOutput:
    @pytest.fixture()
    def output(self):
        output = FFmpegVersionOutput(["/usr/bin/ffmpeg", "-version"], timeout=1.0)
        assert output.command == ["/usr/bin/ffmpeg", "-version"]
        assert output.timeout == 1.0
        assert output.output == []
        assert output.version is None

        return output

    def test_success(self, output: FFmpegVersionOutput):
        output.onstdout(0, "ffmpeg version 0.0.0 suffix")
        assert output.output == ["ffmpeg version 0.0.0 suffix"]
        assert output.version == "0.0.0"

        output.onstdout(1, "foo")
        output.onstdout(2, "bar")
        assert output.output == ["ffmpeg version 0.0.0 suffix", "foo", "bar"]
        assert output.version == "0.0.0"

        assert output.onexit(0)

    def test_failure_stdout(self, output: FFmpegVersionOutput):
        output.onstdout(0, "invalid")
        assert output.output == []
        assert output.version is None
        assert not output.onexit(0)

    def test_failure_exitcode(self, output: FFmpegVersionOutput):
        output.onstdout(0, "ffmpeg version 0.0.0 suffix")
        assert output.output == ["ffmpeg version 0.0.0 suffix"]
        assert output.version == "0.0.0"
        assert not output.onexit(1)


class TestOpen:
    FFMPEG_ARGS_DEFAULT_BASE = ["-y", "-nostats"]
    FFMPEG_ARGS_DEFAULT_LOGLEVEL = ["-loglevel", "info"]
    FFMPEG_ARGS_DEFAULT_CODECS = ["-c:v", FFMPEGMuxer.DEFAULT_VIDEO_CODEC, "-c:a", FFMPEGMuxer.DEFAULT_AUDIO_CODEC]
    FFMPEG_ARGS_DEFAULT_FORMAT = ["-f", FFMPEGMuxer.DEFAULT_OUTPUT_FORMAT]
    FFMPEG_ARGS_DEFAULT_OUTPUT = ["pipe:1"]

    @pytest.fixture(autouse=True)
    def which(self):
        with patch("streamlink.stream.ffmpegmux.which", return_value="ffmpeg") as mock:
            yield mock

    @pytest.fixture()
    def popen(self):
        with patch("subprocess.Popen") as mock:
            yield mock

    @pytest.mark.parametrize(
        ("options", "muxer_args", "expected"),
        [
            pytest.param(
                {},
                {},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="default",
            ),
            pytest.param(
                {},
                {"loglevel": "verbose"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    "-loglevel",
                    "verbose",
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="loglevel",
            ),
            pytest.param(
                {"ffmpeg-loglevel": "error"},
                {"loglevel": "verbose"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    "-loglevel",
                    "error",
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="loglevel-user-override",
            ),
            pytest.param(
                {},
                {"format": "mpegts"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-f",
                    "mpegts",
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="format",
            ),
            pytest.param(
                {"ffmpeg-fout": "avi"},
                {"format": "mpegts"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-f",
                    "avi",
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="format-user-override",
            ),
            pytest.param(
                {},
                {"copyts": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts",
            ),
            pytest.param(
                {"ffmpeg-copyts": True},
                {"copyts": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-user-override",
            ),
            pytest.param(
                {"ffmpeg-start-at-zero": False},
                {"copyts": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-disable-session-start-at-zero",
            ),
            pytest.param(
                {"ffmpeg-copyts": True, "ffmpeg-start-at-zero": False},
                {"copyts": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-disable-session-start-at-zero-user-override",
            ),
            pytest.param(
                {"ffmpeg-start-at-zero": True},
                {"copyts": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    "-start_at_zero",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-enable-session-start-at-zero",
            ),
            pytest.param(
                {"ffmpeg-copyts": True, "ffmpeg-start-at-zero": True},
                {"copyts": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    "-start_at_zero",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-enable-session-start-at-zero-user-override",
            ),
            pytest.param(
                {},
                {"copyts": True, "start_at_zero": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-disable-start-at-zero",
            ),
            pytest.param(
                {"ffmpeg-copyts": True},
                {"copyts": False, "start_at_zero": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-disable-start-at-zero-user-override",
            ),
            pytest.param(
                {},
                {"copyts": True, "start_at_zero": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    "-start_at_zero",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-enable-start-at-zero",
            ),
            pytest.param(
                {"ffmpeg-copyts": True},
                {"copyts": False, "start_at_zero": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    "-start_at_zero",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-enable-start-at-zero-user-override",
            ),
            pytest.param(
                {},
                {"vcodec": "avc"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    "avc",
                    "-c:a",
                    FFMPEGMuxer.DEFAULT_AUDIO_CODEC,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="vcodec",
            ),
            pytest.param(
                {"ffmpeg-video-transcode": "divx"},
                {"vcodec": "avc"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    "divx",
                    "-c:a",
                    FFMPEGMuxer.DEFAULT_AUDIO_CODEC,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="vcodec-user-override",
            ),
            pytest.param(
                {},
                {"acodec": "mp3"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    FFMPEGMuxer.DEFAULT_VIDEO_CODEC,
                    "-c:a",
                    "mp3",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="acodec",
            ),
            pytest.param(
                {"ffmpeg-audio-transcode": "ogg"},
                {"acodec": "mp3"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    FFMPEGMuxer.DEFAULT_VIDEO_CODEC,
                    "-c:a",
                    "ogg",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="acodec-user-override",
            ),
            pytest.param(
                {},
                {"vcodec": "avc", "acodec": "mp3"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    "avc",
                    "-c:a",
                    "mp3",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="vcodec-acodec",
            ),
            pytest.param(
                {"ffmpeg-video-transcode": "divx", "ffmpeg-audio-transcode": "ogg"},
                {"vcodec": "avc", "acodec": "mp3"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    "divx",
                    "-c:a",
                    "ogg",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="vcodec-acodec-user-override",
            ),
            pytest.param(
                {},
                {"maps": ["test", "test2"]},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-map",
                    "test",
                    "-map",
                    "test2",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="maps",
            ),
            pytest.param(
                {},
                {"metadata": {"s:a:0": ["language=eng"]}},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-metadata:s:a:0",
                    "language=eng",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="metadata-stream-audio",
            ),
            pytest.param(
                {},
                {"metadata": {None: ["title=test"]}},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-metadata",
                    "title=test",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="metadata-title",
            ),
        ],
    )
    def test_ffmpeg_args(
        self,
        session: Streamlink,
        popen: Mock,
        options: dict,
        muxer_args: dict,
        expected: list,
    ):
        session.options.update(options)
        streamio = FFMPEGMuxer(session, **muxer_args)

        streamio.open()
        assert popen.call_args_list == [
            call(
                ["ffmpeg", *expected],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            ),
        ]

        streamio.close()

    def test_stderr(self, session: Streamlink, popen: Mock):
        session.options.update({"ffmpeg-verbose": True})
        with patch("streamlink.stream.ffmpegmux.sys.stderr") as mock_stderr:
            streamio = FFMPEGMuxer(session)

            streamio.open()
            assert popen.call_args_list[0][1]["stderr"] is mock_stderr

            streamio.close()
            assert mock_stderr.close.call_count == 0

    @pytest.mark.parametrize(
        ("options", "side_effect"),
        [
            pytest.param({"ffmpeg-verbose-path": "foo"}, None, id="verbose-path"),
            pytest.param({"ffmpeg-verbose-path": "foo", "ffmpeg-verbose": True}, None, id="verbose-path priority"),
            pytest.param({"ffmpeg-verbose-path": "foo"}, OSError, id="OSError on close"),
        ],
    )
    def test_stderr_path(self, session: Streamlink, popen: Mock, options: dict, side_effect: type[Exception] | None):
        session.options.update(options)
        with patch("streamlink.stream.ffmpegmux.Path") as mock_path:
            file: Mock = mock_path("foo").expanduser().open("w")
            file.close.side_effect = side_effect
            streamio = FFMPEGMuxer(session)

            streamio.open()
            assert popen.call_args_list[0][1]["stderr"] is file
            assert file.close.call_count == 0

            streamio.close()
            assert file.close.call_count == 1



import pytest
from lxml.etree import Element

from streamlink.exceptions import PluginError
from streamlink.plugin.api import validate
from streamlink.plugin.api.validate import xml_element
from streamlink.utils.parse import parse_html, parse_json, parse_qsd, parse_xml


class TestUtilsParse:
    def test_parse_json(self):
        assert parse_json("{}") == {}
        assert parse_json('{"test": 1}') == {"test": 1}
        assert parse_json('{"test": 1}', schema=validate.Schema({"test": 1})) == {"test": 1}
        with pytest.raises(PluginError):
            parse_json("""{"test: 1}""")
        with pytest.raises(SyntaxError):
            parse_json("""{"test: 1}""", exception=SyntaxError)
        with pytest.raises(PluginError):
            parse_json("""{"test: 1}""" * 10)

    def test_parse_xml(self):
        expected = Element("test", {"foo": "bar"})
        actual = parse_xml("""<test foo="bar"/>""", ignore_ns=True)
        assert actual.tag == expected.tag
        assert actual.attrib == expected.attrib

    def test_parse_xml_ns_ignore(self):
        expected = Element("test", {"foo": "bar"})
        actual = parse_xml("""<test foo="bar" xmlns="foo:bar"/>""", ignore_ns=True)
        assert actual.tag == expected.tag
        assert actual.attrib == expected.attrib

        actual = parse_xml("""<test	foo="bar"	xmlns="foo:bar"/>""", ignore_ns=True)
        assert actual.tag == expected.tag
        assert actual.attrib == expected.attrib

        actual = parse_xml("""<test\nfoo="bar"\nxmlns="foo:bar"/>""", ignore_ns=True)
        assert actual.tag == expected.tag
        assert actual.attrib == expected.attrib

    def test_parse_xml_ns(self):
        expected = Element("{foo:bar}test", {"foo": "bar"})
        actual = parse_xml("""<h:test foo="bar" xmlns:h="foo:bar"/>""")
        assert actual.tag == expected.tag
        assert actual.attrib == expected.attrib

    def test_parse_xml_fail(self):
        with pytest.raises(PluginError):
            parse_xml("1" * 1000)
        with pytest.raises(SyntaxError):
            parse_xml("1" * 1000, exception=SyntaxError)

    def test_parse_xml_validate(self):
        expected = Element("test", {"foo": "bar"})
        actual = parse_xml(
            """<test foo="bar"/>""",
            schema=validate.Schema(xml_element(tag="test", attrib={"foo": str})),
        )
        assert actual.tag == expected.tag
        assert actual.attrib == expected.attrib

    def test_parse_xml_entities_fail(self):
        with pytest.raises(PluginError):
            parse_xml("""<test foo="bar &"/>""")

    def test_parse_xml_entities(self):
        expected = Element("test", {"foo": "bar &"})
        actual = parse_xml(
            """<test foo="bar &"/>""",
            schema=validate.Schema(xml_element(tag="test", attrib={"foo": str})),
            invalid_char_entities=True,
        )
        assert actual.tag == expected.tag
        assert actual.attrib == expected.attrib

    @pytest.mark.parametrize(
        ("content", "expected"),
        [
            pytest.param(
                """<?xml version="1.0" encoding="UTF-8"?><test>ä</test>""",
                "ä",
                id="string-utf-8",
            ),
            pytest.param(
                """<test>ä</test>""",
                "ä",
                id="string-unknown",
            ),
            pytest.param(
                b"""<?xml version="1.0" encoding="UTF-8"?><test>\xc3\xa4</test>""",
                "ä",
                id="bytes-utf-8",
            ),
            pytest.param(
                b"""<?xml version="1.0" encoding="ISO-8859-1"?><test>\xe4</test>""",
                "ä",
                id="bytes-iso-8859-1",
            ),
            pytest.param(
                b"""<test>\xc3\xa4</test>""",
                "ä",
                id="bytes-unknown",
            ),
        ],
    )
    def test_parse_xml_encoding(self, content, expected):
        tree = parse_xml(content)
        assert tree.xpath(".//text()") == [expected]

    @pytest.mark.parametrize(
        ("content", "expected"),
        [
            pytest.param(
                """<!DOCTYPE html><html><head><meta charset="utf-8"/></head><body>ä</body></html>""",
                "ä",
                id="string-utf-8",
            ),
            pytest.param(
                """<!DOCTYPE html><html><body>ä</body></html>""",
                "ä",
                id="string-unknown",
            ),
            pytest.param(
                b"""<!DOCTYPE html><html><head><meta charset="utf-8"/></head><body>\xc3\xa4</body></html>""",
                "ä",
                id="bytes-utf-8",
            ),
            pytest.param(
                b"""<!DOCTYPE html><html><head><meta charset="ISO-8859-1"/></head><body>\xe4</body></html>""",
                "ä",
                id="bytes-iso-8859-1",
            ),
            pytest.param(
                b"""<!DOCTYPE html><html><body>\xc3\xa4</body></html>""",
                "Ã¤",
                id="bytes-unknown",
            ),
        ],
    )
    def test_parse_html_encoding(self, content, expected):
        tree = parse_html(content)
        assert tree.xpath(".//body/text()") == [expected]

    @pytest.mark.parametrize(
        ("content", "expected"),
        [
            pytest.param(
                """<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE html><html><body>ä?></body></html>""",
                "ä?>",
                id="string",
            ),
            pytest.param(
                b"""<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE html><html><body>\xc3\xa4?></body></html>""",
                "ä?>",
                id="bytes-utf-8",
            ),
            pytest.param(
                b"""<?xml version="1.0" encoding="ISO-8859-1"?><!DOCTYPE html><html><body>\xe4?></body></html>""",
                "ä?>",
                id="bytes-iso-8859-1",
            ),
            pytest.param(
                b"""<?xml version="1.0"?><!DOCTYPE html><html><body>\xc3\xa4?></body></html>""",
                "ä?>",
                id="bytes-unknown",
            ),
        ],
    )
    def test_parse_html_xhtml5(self, content, expected):
        tree = parse_html(content)
        assert tree.xpath(".//body/text()") == [expected]

    def test_parse_qsd(self):
        assert parse_qsd("test=1&foo=bar", schema=validate.Schema({"test": str, "foo": "bar"})) == {"test": "1", "foo": "bar"}



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from io import BufferedRandom, BufferedWriter
from pathlib import Path

import pytest

from streamlink_cli.output import FileOutput


@contextmanager
def _create_fd(root: Path, name: str) -> Iterator[BufferedRandom]:
    fd = (root / name).open("w+b")
    try:
        yield fd
    finally:
        fd.close()


@pytest.fixture()
def fd(tmp_path: Path):
    with _create_fd(tmp_path, "file") as fd:
        yield fd


@pytest.fixture(autouse=True)
def fake_stdout(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    # can't use in-memory io.BytesIO, since fd.fileno() is called on Windows
    with _create_fd(tmp_path, "stdout") as fd:
        monkeypatch.setattr("streamlink_cli.output.file.stdout", fd)
        yield fd


def test_early_close(tmp_path: Path, fd: BufferedRandom):
    filename = tmp_path / "foo" / "bar"
    fo = FileOutput(filename=filename, record=FileOutput(fd=fd))
    assert isinstance(fo.record, FileOutput)
    assert not fo.opened
    assert not fo.record.opened
    assert not filename.exists()

    fo.close()
    fo.record.close()


def test_early_write(tmp_path: Path):
    filename = tmp_path / "foo" / "bar"
    fo = FileOutput(filename=filename)

    assert not fo.opened
    assert not filename.exists()
    with pytest.raises(OSError, match=r"^Output is not opened$"):
        fo.write(b"foo")


def test_open_write_close(tmp_path: Path, fd: BufferedRandom):
    filename = tmp_path / "foo" / "bar"
    fo = FileOutput(filename=filename, record=FileOutput(fd=fd))
    assert fo.fd is None
    assert isinstance(fo.record, FileOutput)

    fo.open()
    assert fo.opened
    assert fo.record.opened
    assert filename.parent.is_dir()
    assert filename.is_file()
    assert isinstance(fo.fd, BufferedWriter)
    assert isinstance(fo.record.fd, BufferedRandom)

    fo.write(b"foo")
    fo.write(b"bar")
    fo.write(b"baz")
    fo.fd.flush()
    fo.record.fd.flush()
    assert filename.read_bytes() == b"foobarbaz"
    fo.record.fd.seek(0)
    assert fo.record.fd.read() == b"foobarbaz"

    fo.close()
    assert not fo.opened
    assert not fo.record.opened
    assert fo.fd.closed
    assert fo.record.fd.closed


def test_write_stdout(fake_stdout: BufferedRandom):
    fo = FileOutput(fd=fake_stdout)
    assert fo.fd is fake_stdout
    assert fo.filename is None
    assert fo.record is None

    fo.open()
    assert fo.opened

    fo.write(b"foo")
    fo.write(b"bar")
    fo.write(b"baz")
    fo.fd.seek(0)
    assert fo.fd.read() == b"foobarbaz"

    fo.close()
    assert not fo.opened
    assert not fo.fd.closed



from __future__ import annotations

import json
import re
from typing import Iterator
from unittest.mock import Mock

import pytest

import streamlink_cli.main
from streamlink.exceptions import FatalPluginError, PluginError
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.session import Streamlink
from streamlink.stream.stream import Stream


STREAMS = {
    "audio": "mock://stream/audio",
    "720p": "mock://stream/720p",
    "1080p": "mock://stream/1080p",
    "worst": "mock://stream/720p",
    "best": "mock://stream/1080p",
}
STREAMS_MULTIVARIANT = "mock://stream"


class FakeStream(Stream):
    __shortname__ = "fake"

    def __init__(self, session: Streamlink, url: str | None, manifest_url: str | None):
        super().__init__(session)
        self.url = url
        self.manifest_url = manifest_url

    def __json__(self):  # noqa: PLW3201
        return {
            "type": self.__shortname__,
            "url": self.url,
            "manifest_url": self.manifest_url,
        }

    def to_url(self):
        return self.url if self.url else super().to_url()

    def to_manifest_url(self):
        return self.manifest_url if self.manifest_url else super().to_url()


@pytest.fixture(autouse=True)
def _stream_output(monkeypatch: pytest.MonkeyPatch):
    mock_output_stream = Mock()
    mock_output_stream_http = Mock()
    mock_output_stream_passthrough = Mock()
    monkeypatch.setattr(streamlink_cli.main, "output_stream", mock_output_stream)
    monkeypatch.setattr(streamlink_cli.main, "output_stream_http", mock_output_stream_http)
    monkeypatch.setattr(streamlink_cli.main, "output_stream_passthrough", mock_output_stream_passthrough)


@pytest.fixture(autouse=True)
def streams(request: pytest.FixtureRequest, session: Streamlink):
    params = getattr(request, "param", [{}])
    params = params if isinstance(params, list) else [params]

    def streams_generator():
        for param in params:
            if exc := param.get("exc"):
                yield exc
                continue

            streams = param.get("streams", {})
            to_url = param.get("to_url", True)
            to_manifest_url = param.get("to_manifest_url", True)

            yield {
                name: FakeStream(
                    session,
                    url=url if to_url else None,
                    manifest_url=STREAMS_MULTIVARIANT if to_manifest_url else None,
                )
                for name, url in (STREAMS if streams is True else streams).items()
            }

    return streams_generator()


@pytest.fixture(autouse=True)
def plugin(session: Streamlink, streams: Iterator[dict[str, FakeStream]]):
    @pluginmatcher(re.compile(r"https?://plugin"))
    class FakePlugin(Plugin):
        __module__ = "streamlink.plugins.plugin"

        id = "ID"
        author = "AUTHOR"
        category = "CATEGORY"
        title = "TITLE"

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.logger.info("plugin log message")

        def _get_streams(self):
            item = next(streams, None)
            if isinstance(item, BaseException):
                raise item

            return item

    session.plugins.update({"plugin": FakePlugin})


@pytest.mark.parametrize(
    ("argv", "streams", "exit_code", "stdout"),
    [
        pytest.param(
            ["doesnotexist"],
            {},
            1,
            "error: No plugin can handle URL: doesnotexist\n",
            id="no-plugin",
        ),
        pytest.param(
            ["plugin"],
            {"exc": PluginError("Error while fetching streams")},
            1,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "error: Error while fetching streams\n"
            ),
            id="fetch-streams-exception",
        ),
        pytest.param(
            ["plugin"],
            {},
            1,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "error: No playable streams found on this URL: plugin\n"
            ),
            id="no-streams",
        ),
        pytest.param(
            ["plugin"],
            {"streams": True},
            0,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "Available streams: audio, 720p (worst), 1080p (best)\n"
            ),
            id="streams-selection-none",
        ),
        pytest.param(
            ["plugin", "one,two,three"],
            {"streams": True},
            1,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "error: The specified stream(s) 'one, two, three' could not be found.\n"
                + "       Available streams: audio, 720p (worst), 1080p (best)\n"
            ),
            id="streams-selection-invalid",
        ),
        pytest.param(
            ["plugin", "best"],
            {"streams": True},
            0,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "[cli][info] Available streams: audio, 720p (worst), 1080p (best)\n"
                + "[cli][info] Opening stream: 1080p (fake)\n"
            ),
            id="streams-selection-best",
        ),
        pytest.param(
            ["--stream-url", "plugin"],
            {"streams": True},
            0,
            f"{STREAMS_MULTIVARIANT}\n",
            id="stream-url-selection-none",
        ),
        pytest.param(
            ["--stream-url", "plugin"],
            {"streams": True, "to_manifest_url": False},
            1,
            "error: The stream specified cannot be translated to a URL\n",
            id="stream-url-selection-none-no-multivariant",
        ),
        pytest.param(
            ["--stream-url", "plugin", "best"],
            {"streams": True},
            0,
            f"{STREAMS['best']}\n",
            id="stream-url-selection-best",
        ),
        pytest.param(
            ["--stream-url", "plugin", "best"],
            {"streams": True, "to_url": False},
            1,
            "error: The stream specified cannot be translated to a URL\n",
            id="stream-url-selection-best-no-url",
        ),
        pytest.param(
            ["--json", "plugin"],
            {"streams": True},
            0,
            json.dumps(
                {
                    "plugin": "plugin",
                    "metadata": {
                        "id": "ID",
                        "author": "AUTHOR",
                        "category": "CATEGORY",
                        "title": "TITLE",
                    },
                    "streams": {
                        "audio": {
                            "type": "fake",
                            "url": "mock://stream/audio",
                            "manifest_url": "mock://stream",
                        },
                        "worst": {
                            "type": "fake",
                            "url": "mock://stream/720p",
                            "manifest_url": "mock://stream",
                        },
                        "best": {
                            "type": "fake",
                            "url": "mock://stream/1080p",
                            "manifest_url": "mock://stream",
                        },
                        "720p": {
                            "type": "fake",
                            "url": "mock://stream/720p",
                            "manifest_url": "mock://stream",
                        },
                        "1080p": {
                            "type": "fake",
                            "url": "mock://stream/1080p",
                            "manifest_url": "mock://stream",
                        },
                    },
                },
                indent=2,
                separators=(",", ": "),
            )
            + "\n",
            id="json-selection-none",
        ),
        pytest.param(
            ["--json", "plugin", "one,two,three"],
            {"streams": True},
            1,
            json.dumps(
                {
                    "plugin": "plugin",
                    "metadata": {
                        "id": "ID",
                        "author": "AUTHOR",
                        "category": "CATEGORY",
                        "title": "TITLE",
                    },
                    "streams": {
                        "audio": {
                            "type": "fake",
                            "url": "mock://stream/audio",
                            "manifest_url": "mock://stream",
                        },
                        "worst": {
                            "type": "fake",
                            "url": "mock://stream/720p",
                            "manifest_url": "mock://stream",
                        },
                        "best": {
                            "type": "fake",
                            "url": "mock://stream/1080p",
                            "manifest_url": "mock://stream",
                        },
                        "720p": {
                            "type": "fake",
                            "url": "mock://stream/720p",
                            "manifest_url": "mock://stream",
                        },
                        "1080p": {
                            "type": "fake",
                            "url": "mock://stream/1080p",
                            "manifest_url": "mock://stream",
                        },
                    },
                    "error": "The specified stream(s) 'one, two, three' could not be found",
                },
                indent=2,
                separators=(",", ": "),
            )
            + "\n",
            id="json-selection-invalid",
        ),
        pytest.param(
            ["--json", "plugin", "best"],
            {"streams": True},
            0,
            json.dumps(
                {
                    "type": "fake",
                    "url": "mock://stream/1080p",
                    "manifest_url": "mock://stream",
                    "metadata": {
                        "id": "ID",
                        "author": "AUTHOR",
                        "category": "CATEGORY",
                        "title": "TITLE",
                    },
                },
                indent=2,
                separators=(",", ": "),
            )
            + "\n",
            id="json-selection-best",
        ),
    ],
    indirect=["argv", "streams"],
)
def test_handle_url_text_output(
    capsys: pytest.CaptureFixture[str],
    argv: list,
    streams: Iterator[dict[str, FakeStream]],
    exit_code: int,
    stdout: str,
):
    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()

    assert exc_info.value.code == exit_code
    out, err = capsys.readouterr()
    assert out == stdout
    assert err == ""


@pytest.mark.parametrize(
    ("argv", "streams", "exit_code", "retries", "stdout"),
    [
        pytest.param(
            ["plugin", "best"],
            [],
            1,
            0,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "error: No playable streams found on this URL: plugin\n"
            ),
            id="no-retries-implicit",
        ),
        pytest.param(
            ["plugin", "best", "--retry-streams=0", "--retry-max=0"],
            [],
            1,
            0,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "error: No playable streams found on this URL: plugin\n"
            ),
            id="no-retries-explicit",
        ),
        pytest.param(
            ["plugin", "best", "--retry-max=5"],
            [],
            1,
            5,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "[cli][info] Waiting for streams, retrying every 1 second(s)\n"
                + "error: No playable streams found on this URL: plugin\n"
            ),
            id="no-streams",
        ),
        pytest.param(
            ["plugin", "best", "--retry-max=5"],
            [{"streams": True}],
            0,
            0,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "[cli][info] Available streams: audio, 720p (worst), 1080p (best)\n"
                + "[cli][info] Opening stream: 1080p (fake)\n"
            ),
            id="success-on-first-attempt",
        ),
        pytest.param(
            ["plugin", "best", "--retry-streams=3", "--retry-max=5"],
            [{}, {}, {"streams": True}],
            0,
            2,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "[cli][info] Waiting for streams, retrying every 3.0 second(s)\n"
                + "[cli][info] Available streams: audio, 720p (worst), 1080p (best)\n"
                + "[cli][info] Opening stream: 1080p (fake)\n"
            ),
            id="success-on-third-attempt",
        ),
        pytest.param(
            ["plugin", "best", "--retry-streams=3", "--retry-max=5"],
            [{"exc": PluginError("failure")}, {"streams": True}],
            0,
            1,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "[cli][error] failure\n"
                + "[cli][info] Waiting for streams, retrying every 3.0 second(s)\n"
                + "[cli][info] Available streams: audio, 720p (worst), 1080p (best)\n"
                + "[cli][info] Opening stream: 1080p (fake)\n"
            ),
            id="success-with-plugin-error-on-first-attempt",
        ),
        pytest.param(
            ["plugin", "best", "--retry-streams=3", "--retry-max=5"],
            [{}, {"exc": PluginError("failure")}, {"streams": True}],
            0,
            2,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "[cli][info] Waiting for streams, retrying every 3.0 second(s)\n"
                + "[cli][error] failure\n"
                + "[cli][info] Available streams: audio, 720p (worst), 1080p (best)\n"
                + "[cli][info] Opening stream: 1080p (fake)\n"
            ),
            id="success-with-plugin-error-on-second-attempt",
        ),
        pytest.param(
            ["plugin", "best", "--retry-streams=3"],
            [{} for _ in range(20)] + [{"streams": True}],
            0,
            20,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "[cli][info] Waiting for streams, retrying every 3.0 second(s)\n"
                + "[cli][info] Available streams: audio, 720p (worst), 1080p (best)\n"
                + "[cli][info] Opening stream: 1080p (fake)\n"
            ),
            id="success-no-max-attempts",
        ),
        pytest.param(
            ["plugin", "best", "--retry-max=5"],
            [{"exc": FatalPluginError("fatal")}, {"streams": True}],
            1,
            0,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "error: fatal\n"
            ),
            id="fatal-plugin-error-on-first-attempt",
        ),
        pytest.param(
            ["plugin", "best", "--retry-max=5"],
            [{}, {"exc": FatalPluginError("fatal")}, {"streams": True}],
            1,
            1,
            (
                "[cli][info] Found matching plugin plugin for URL plugin\n"
                + "[plugins.plugin][info] plugin log message\n"
                + "[cli][info] Waiting for streams, retrying every 1 second(s)\n"
                + "error: fatal\n"
            ),
            id="fatal-plugin-error-on-second-attempt",
        ),
    ],
    indirect=["argv", "streams"],
)  # fmt: skip
def test_handle_url_retry(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    argv: list,
    streams: Iterator[dict[str, FakeStream]],
    exit_code: int,
    retries: int,
    stdout: list[str],
):
    mock_sleep = Mock()
    monkeypatch.setattr("streamlink_cli.main.sleep", mock_sleep)

    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()

    assert exc_info.value.code == exit_code
    assert mock_sleep.call_count == retries

    out, err = capsys.readouterr()
    assert out == stdout
    assert err == ""


@pytest.mark.parametrize("argv", [pytest.param(["plugin", "best"], id="argv")], indirect=["argv"])
class TestHandleUrlKeyboardInterruptAndCleanup:
    @pytest.fixture(autouse=True)
    def handle_url(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        mock_handle_url = Mock()
        if getattr(request, "param", False):
            mock_handle_url.side_effect = KeyboardInterrupt
        monkeypatch.setattr(streamlink_cli.main, "handle_url", mock_handle_url)

    @staticmethod
    def _mock(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch, attr: str):
        param = getattr(request, "param", {})
        if not param.get("initialized", False):
            monkeypatch.setattr(streamlink_cli.main, attr, None)
            yield
        else:
            mock = Mock()
            monkeypatch.setattr(streamlink_cli.main, attr, mock)
            if param.get("close_raises", False):
                mock.close.side_effect = KeyboardInterrupt

            yield mock
            assert mock.close.call_count == 1

    @pytest.fixture(autouse=True)
    def output(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        yield from self._mock(request, monkeypatch, "output")

    @pytest.fixture(autouse=True)
    def stream_fd(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        yield from self._mock(request, monkeypatch, "stream_fd")

    @pytest.mark.parametrize(
        ("handle_url", "output", "stream_fd", "exit_code", "stdout"),
        [
            pytest.param(
                False,
                {},
                {},
                0,
                "",
                id="no-keyboardinterrupt",
            ),
            pytest.param(
                True,
                {},
                {},
                130,
                "Interrupted! Exiting...\n",
                id="no-output",
            ),
            pytest.param(
                True,
                {"initialized": True},
                {},
                130,
                "Interrupted! Exiting...\n",
                id="output",
            ),
            pytest.param(
                True,
                {"initialized": True},
                {"initialized": True},
                130,
                "Interrupted! Exiting...\n[cli][info] Closing currently open stream...\n",
                id="output-streamfd",
            ),
            pytest.param(
                True,
                {"initialized": True, "close_raises": True},
                {"initialized": True},
                130,
                "Interrupted! Exiting...\n[cli][info] Closing currently open stream...\n",
                id="output-streamfd-outputclose-interrupted",
            ),
            pytest.param(
                True,
                {"initialized": True},
                {"initialized": True, "close_raises": True},
                130,
                "Interrupted! Exiting...\n[cli][info] Closing currently open stream...\n",
                id="output-streamfd-streamfdclose-interrupted",
            ),
        ],
        indirect=["handle_url", "output", "stream_fd"],
    )
    def test_handle_url(
        self,
        capsys: pytest.CaptureFixture[str],
        argv: list,
        handle_url: None,
        output: Mock,
        stream_fd: Mock,
        exit_code: int,
        stdout: str,
    ):
        # noinspection PyTypeChecker
        with pytest.raises((SystemExit, KeyboardInterrupt)) as exc_info:
            streamlink_cli.main.main()

        assert isinstance(exc_info.value, SystemExit)
        assert exc_info.value.code == exit_code
        out, err = capsys.readouterr()
        assert out == stdout
        assert err == ""



from __future__ import annotations

import shlex
from typing import Any

from setuptools import build_meta as _build_meta

# re-export everything from `setuptools.build_meta`, so that we don't have to worry about any hooks which we don't override
# https://peps.python.org/pep-0517/
# https://peps.python.org/pep-0660/
# noinspection PyUnresolvedReferences
from setuptools.build_meta import *  # noqa: F403
from setuptools.command.egg_info import egg_info as _egg_info


# ----


def get_requires_for_build_wheel(  # type: ignore[no-redef]
    config_settings: dict | None = None,
) -> list[str]:  # pragma: no cover
    # Streamlink publishes three wheels on PyPI: the generic "any" wheel, the "win32" wheel and the "win-amd64" wheel:
    # The Windows-wheels are special, because they include a "gui_scripts" entry point, which is used by the `pip` frontend
    # to generate the "streamlinkw" launcher, which doesn't open a terminal window when launching it from a GUI application.
    #
    # In order to build these special Windows-wheels, the `--plat-name=...` CLI argument needs to get passed
    # to the `bdist_wheel` setuptools command (provided by the `wheel` package). With the introduction of PEP517 however,
    # setuptools's CLI is not used directly anymore, and instead, we have to build the wheels using the `build` package
    # and the `--wheel --config-setting=--build-option=...` args, which are then forwarded to setuptools via the PEP517 hooks.
    #
    # Since `build==1.0.0` though, the `--config-setting` data now gets passed to all build-backend hooks involved,
    # even `get_requires_for_build_wheel()`. This results in the build failing, as it executes setuptools's `egg_info` command,
    # which doesn't accept the `--plat-name` argument or any other `bdist_wheel` command options.
    #
    # As a consequence of this, we need to override the build-backend and the `get_requires_for_build_wheel()` hook, so that
    # we can filter out arguments which are not relevant to the `egg_info` command.
    _filter_cmd_option_args(
        config_settings,
        "--build-option",
        # types-setuptools-70.3.0.20240710 has the wrong type for
        #   setuptools.command.egg_info.user_options (setuptools._distutils.cmd.Command.user_options)
        _egg_info.user_options,  # type: ignore[arg-type]
    )

    return _build_meta.get_requires_for_build_wheel(config_settings)


# ----


def _filter_cmd_option_args(
    config_settings: dict | None,
    key: str,
    # https://github.com/pypa/setuptools/blob/v71.1.0/setuptools/_distutils/fancy_getopt.py#L47-L54
    # https://github.com/pypa/setuptools/blob/v71.1.0/setuptools/_distutils/fancy_getopt.py#L152-L156
    options: list[tuple[str, str | None, str] | tuple[str, str | None, str, Any]],
) -> None:
    """Filter out args which are not recognized by a specific command and its options"""

    if not config_settings or not config_settings.get(key):
        return

    parsed = shlex.split(config_settings[key])

    result = []
    val_next = False
    for item in parsed:
        if val_next:
            val_next = False
            result.append(item)
            continue
        full: str
        shorthand: str | None
        for full, shorthand, *_ in options:
            is_boolean = full[-1] != "="
            is_shorthand = shorthand is not None and item == f"-{shorthand}"
            if not is_boolean and (is_shorthand or item == f"--{full[:-1]}"):
                val_next = True
                result.append(item)
                break
            if (
                is_boolean and (is_shorthand or item == f"--{full}")
                or not is_boolean and item.startswith(f"--{full}")
            ):  # fmt: skip
                result.append(item)
                break

    if result:
        config_settings[key] = shlex.join(result)
    else:
        del config_settings[key]



from streamlink.plugins.tv5monde import TV5Monde
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTV5Monde(PluginCanHandleUrl):
    __plugin__ = TV5Monde

    should_match = [
        "https://live.tv5monde.com/fbs.html",
        "https://europe.tv5monde.com/fr/direct",
        "https://maghreb-orient.tv5monde.com/fr/direct",
        "https://revoir.tv5monde.com/toutes-les-videos/documentaires/sauver-notre-dame-sauver-notre-dame-14-04-2020",
        "https://information.tv5monde.com/video/la-diplomatie-francaise-est-elle-en-crise",
        "https://afrique.tv5monde.com/videos/afrykas-et-la-boite-magique",
        "https://www.tivi5mondeplus.com/conte-nous/episode-25",
    ]



from streamlink.plugins.abematv import AbemaTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlAbemaTV(PluginCanHandleUrl):
    __plugin__ = AbemaTV

    should_match_groups = [
        (("onair", "https://abema.tv/now-on-air/abema-news"), {"onair": "abema-news"}),
        (("onair", "https://abema.tv/now-on-air/abema-news?a=b&c=d"), {"onair": "abema-news"}),
        (("episode", "https://abema.tv/video/episode/90-1053_s99_p12"), {"episode": "90-1053_s99_p12"}),
        (("episode", "https://abema.tv/video/episode/90-1053_s99_p12?a=b&c=d"), {"episode": "90-1053_s99_p12"}),
        (("slots", "https://abema.tv/channels/everybody-anime/slots/FJcUsdYjTk1rAb"), {"slots": "FJcUsdYjTk1rAb"}),
        (("slots", "https://abema.tv/channels/abema-anime/slots/9rTULtcJFiFmM9?a=b"), {"slots": "9rTULtcJFiFmM9"}),
    ]

    should_not_match = [
        "https://www.abema.tv/now-on-air/abema-news",
        "https://www.abema.tv/now-on-air/",
        "https://abema.tv/timetable",
        "https://abema.tv/video",
        "https://abema.tv/video/title/13-47",
        "https://abema.tv/video/title/13-47?a=b",
    ]



from streamlink.plugins.radiko import Radiko
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRadiko(PluginCanHandleUrl):
    __plugin__ = Radiko

    should_match = [
        "https://radiko.jp/#!/live/QRR",
        "https://radiko.jp/#!/ts/YFM/20201206010000",
        "http://radiko.jp/#!/live/QRR",
        "http://radiko.jp/live/QRR",
        "http://radiko.jp/#!/ts/QRR/20200308180000",
        "http://radiko.jp/ts/QRR/20200308180000",
    ]



from streamlink.plugins.piaulizaportal import PIAULIZAPortal
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlPIAULIZAPortal(PluginCanHandleUrl):
    __plugin__ = PIAULIZAPortal

    should_match_groups = [
        (
            "https://ulizaportal.jp/pages/005f18b7-e810-5618-cb82-0987c5755d44",
            {"id": "005f18b7-e810-5618-cb82-0987c5755d44"},
        ),
        (
            "https://ulizaportal.jp/pages/005e1b23-fe93-5780-19a0-98e917cc4b7d"
            + "?expires=4102412400&signature=f422a993b683e1068f946caf406d211c17d1ef17da8bef3df4a519502155aa91&version=1",
            {"id": "005e1b23-fe93-5780-19a0-98e917cc4b7d"},
        ),
    ]

    should_not_match = [
        "https://ulizaportal.jp/pages/",
        "https://ulizaportal.jp/pages/invalid-id",
    ]



import re
from pathlib import Path
from unittest.mock import Mock, call

import pytest

from streamlink.exceptions import StreamlinkDeprecationWarning
from streamlink_cli.compat import stdout
from streamlink_cli.exceptions import StreamlinkCLIError
from streamlink_cli.main import (
    Formatter,
    build_parser,
    create_output,
    setup_args,
)
from streamlink_cli.output import FileOutput, PlayerOutput


ARGS_PLAYER_ENV = "--player-env=VAR1=abc", "--player-env=VAR2=def"


@pytest.fixture(autouse=True)
def argv(argv: list):
    parser = build_parser()
    setup_args(parser)

    return argv


@pytest.fixture(autouse=True)
def _default_stream_metadata(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("streamlink_cli.main.DEFAULT_STREAM_METADATA", {"title": "bar"})


@pytest.fixture()
def formatter():
    return Formatter(
        {
            "author": lambda: "foo",
        },
    )


@pytest.fixture()
def check_file_output(monkeypatch: pytest.MonkeyPatch):
    mock_check_file_output = Mock(side_effect=lambda path, force: path)
    monkeypatch.setattr("streamlink_cli.main.check_file_output", mock_check_file_output)

    return mock_check_file_output


@pytest.mark.parametrize(
    ("argv", "title"),
    [
        pytest.param(
            ["--player=mpv", "--player-args=--no-border", *ARGS_PLAYER_ENV, "URL"],
            "URL",
            id="title-default",
        ),
        pytest.param(
            ["--player=mpv", "--player-args=--no-border", *ARGS_PLAYER_ENV, "--title={author} - {title}", "URL"],
            "foo - bar",
            id="title-custom",
        ),
    ],
    indirect=["argv"],
)
def test_player(argv: list, formatter: Formatter, title: str):
    output = create_output(formatter)
    assert isinstance(output, PlayerOutput)
    assert output.playerargs.path == Path("mpv")
    assert output.playerargs.args == "--no-border"
    assert output.playerargs.title == title
    assert output.env == {"VAR1": "abc", "VAR2": "def"}
    assert output.record is None


@pytest.mark.parametrize(
    ("argv", "force", "title"),
    [
        pytest.param(
            ["--record=foo", "--player=mpv", *ARGS_PLAYER_ENV, "URL"],
            False,
            "URL",
            id="no-force-title-default",
        ),
        pytest.param(
            ["--record=foo", "--player=mpv", *ARGS_PLAYER_ENV, "--title={author} - {title}", "URL"],
            False,
            "foo - bar",
            id="title-custom",
        ),
        pytest.param(
            ["--record=foo", "--force", "--player=mpv", *ARGS_PLAYER_ENV, "URL"],
            True,
            "URL",
            id="force",
        ),
    ],
    indirect=["argv"],
)
def test_player_record(check_file_output: Mock, formatter: Formatter, argv: list, force: bool, title: str):
    output = create_output(formatter)
    assert check_file_output.call_args_list == [call(Path("foo"), force)]
    assert isinstance(output, PlayerOutput)
    assert output.playerargs.title == title
    assert output.env == {"VAR1": "abc", "VAR2": "def"}
    assert isinstance(output.record, FileOutput)
    assert output.record.filename == Path("foo")
    assert output.record.fd is None
    assert output.record.record is None


@pytest.mark.parametrize(
    "argv",
    [pytest.param(["--record=-", "--player=mpv", *ARGS_PLAYER_ENV, "--title={author} - {title}", "URL"])],
    indirect=["argv"],
)
def test_player_record_stdout(formatter: Formatter, argv: list):
    output = create_output(formatter)
    assert type(output) is PlayerOutput
    assert output.playerargs.title == "foo - bar"
    assert output.env == {"VAR1": "abc", "VAR2": "def"}
    assert type(output.record) is FileOutput
    assert output.record.filename is None
    assert output.record.fd is stdout
    assert output.record.record is None


@pytest.mark.parametrize(
    ("argv", "force"),
    [
        pytest.param(["--output=foo"], False, id="no-force"),
        pytest.param(["--output=foo", "--force"], True, id="force"),
    ],
    indirect=["argv"],
)
def test_output(check_file_output: Mock, formatter: Formatter, argv: list, force: bool):
    output = create_output(formatter)
    assert check_file_output.call_args_list == [call(Path("foo"), force)]
    assert isinstance(output, FileOutput)
    assert output.filename == Path("foo")
    assert output.fd is None
    assert output.record is None


@pytest.mark.parametrize(
    "argv",
    [
        pytest.param(["--stdout"], id="stdout"),
        pytest.param(["--stdout", "--record=-"], id="stdout-and-record-dash"),
        pytest.param(["--output=-"], id="output-dash"),
    ],
    indirect=["argv"],
)
def test_stdout(formatter: Formatter, argv: list):
    output = create_output(formatter)
    assert isinstance(output, FileOutput)
    assert output.filename is None
    assert output.fd is stdout
    assert output.record is None


@pytest.mark.parametrize(
    ("argv", "force", "warnings"),
    [
        pytest.param(
            ["--stdout", "--record=foo"],
            False,
            [],
            id="no-force",
        ),
        pytest.param(
            ["--stdout", "--record=foo", "--force"],
            True,
            [],
            id="force",
        ),
        pytest.param(
            ["--record-and-pipe=foo"],
            False,
            [(StreamlinkDeprecationWarning, "-R/--record-and-pipe=... has been deprecated in favor of --stdout --record=...")],
            id="record-and-pipe-no-force",
        ),
        pytest.param(
            ["--record-and-pipe=foo", "--force"],
            True,
            [(StreamlinkDeprecationWarning, "-R/--record-and-pipe=... has been deprecated in favor of --stdout --record=...")],
            id="record-and-pipe-force",
        ),
    ],
    indirect=["argv"],
)
def test_stdout_record(
    recwarn: pytest.WarningsRecorder,
    check_file_output: Mock,
    formatter: Formatter,
    argv: list,
    force: bool,
    warnings: list,
):
    output = create_output(formatter)
    assert check_file_output.call_args_list == [call(Path("foo"), force)]
    assert isinstance(output, FileOutput)
    assert output.filename is None
    assert output.fd is stdout
    assert isinstance(output.record, FileOutput)
    assert output.record.filename == Path("foo")
    assert output.record.fd is None
    assert output.record.record is None
    assert [(record.category, str(record.message)) for record in recwarn.list] == warnings


@pytest.mark.parametrize(
    ("argv", "errormsg"),
    [
        pytest.param(
            ["--output=foo", "--stdout"],
            "The -o/--output argument is incompatible with -O/--stdout",
            id="output-stdout",
        ),
        pytest.param(
            ["--output=foo", "--record=bar"],
            "The -o/--output argument is incompatible with -r/--record and -R/--record-and-pipe",
            id="output-record",
        ),
        pytest.param(
            ["--output=foo", "--record-and-pipe=bar"],
            "The -o/--output argument is incompatible with -r/--record and -R/--record-and-pipe",
            id="output-record-and-pipe",
        ),
        pytest.param(
            ["--stdout", "--record-and-pipe=bar"],
            "The -O/--stdout argument is incompatible with -R/--record-and-pipe",
            id="stdout-record-and-pipe",
        ),
    ],
    indirect=["argv"],
)
def test_incompatible_options(formatter: Formatter, argv: list, errormsg: str):
    with pytest.raises(StreamlinkCLIError) as excinfo:
        create_output(formatter)
    assert str(excinfo.value) == errormsg
    assert excinfo.value.code == 1


@pytest.mark.parametrize("argv", [pytest.param([], id="default-player")], indirect=["argv"])
def test_no_default_player(formatter: Formatter, argv: list):
    with pytest.raises(StreamlinkCLIError) as excinfo:
        create_output(formatter)
    assert re.search(r"^The default player \(\w+\) does not seem to be installed\.", str(excinfo.value))
    assert excinfo.value.code == 1



from streamlink.plugins.artetv import ArteTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlArteTV(PluginCanHandleUrl):
    __plugin__ = ArteTV

    should_match_groups = [
        # live
        (
            ("live", "https://www.arte.tv/fr/direct"),
            {"language": "fr"},
        ),
        (
            ("live", "https://www.arte.tv/fr/direct/"),
            {"language": "fr"},
        ),
        (
            ("live", "https://www.arte.tv/de/live"),
            {"language": "de"},
        ),
        (
            ("live", "https://www.arte.tv/de/live/"),
            {"language": "de"},
        ),
        # vod
        (
            ("vod", "https://www.arte.tv/de/videos/097372-001-A/mysterium-satoshi-bitcoin-wie-alles-begann-1-6/"),
            {"language": "de", "video_id": "097372-001-A"},
        ),
        (
            ("vod", "https://www.arte.tv/en/videos/097372-001-A/the-satoshi-mystery-the-story-of-bitcoin/"),
            {"language": "en", "video_id": "097372-001-A"},
        ),
        # old vod URLs with redirects
        (
            ("vod", "https://www.arte.tv/guide/de/097372-001-A/mysterium-satoshi-bitcoin-wie-alles-begann-1-6/"),
            {"language": "de", "video_id": "097372-001-A"},
        ),
        (
            ("vod", "https://www.arte.tv/guide/en/097372-001-A/the-satoshi-mystery-the-story-of-bitcoin/"),
            {"language": "en", "video_id": "097372-001-A"},
        ),
    ]

    should_not_match = [
        "https://www.arte.tv/guide/de/live/",
        "https://www.arte.tv/guide/fr/plus7/",
        "https://www.arte.tv/guide/de/plus7/",
        # shouldn't match - playlists without video ids in url
        "https://www.arte.tv/en/videos/RC-014457/the-power-of-forests/",
        "https://www.arte.tv/en/videos/RC-013118/street-art/",
    ]



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from unittest.mock import Mock, call, sentinel

import pytest

import streamlink_cli.main
from streamlink.exceptions import StreamError
from streamlink.stream.stream import Stream
from streamlink_cli.constants import PROGRESS_INTERVAL_NO_STATUS
from streamlink_cli.exceptions import StreamlinkCLIError
from streamlink_cli.main import build_parser, setup_args
from streamlink_cli.output import FileOutput, PlayerOutput


@pytest.fixture(autouse=True)
def argv(argv: list):
    parser = build_parser()
    setup_args(parser)

    return argv


@pytest.fixture()
def caplog(caplog: pytest.LogCaptureFixture):
    caplog.set_level(1, "streamlink.cli")
    return caplog


@pytest.fixture(autouse=True)
def output(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    output = getattr(request, "param", Mock())
    monkeypatch.setattr(streamlink_cli.main, "output", output)
    monkeypatch.setattr(streamlink_cli.main, "create_output", Mock(return_value=output))
    monkeypatch.setattr(output, "open", Mock())

    return output


@pytest.fixture()
def stream(monkeypatch: pytest.MonkeyPatch):
    class FakeStream(Stream):
        __shortname__ = "fake-stream"

        def __str__(self):
            return self.__shortname__

    return FakeStream(Mock())


@pytest.mark.parametrize("argv", [pytest.param(["--retry-open=2"])], indirect=True)
def test_stream_failure_no_output_open(
    monkeypatch: pytest.MonkeyPatch,
    caplog: pytest.LogCaptureFixture,
    argv: list,
    output: Mock,
    stream: Stream,
):
    monkeypatch.setattr(stream, "open", Mock(side_effect=StreamError("failure")))

    with pytest.raises(StreamlinkCLIError) as exc_info:
        streamlink_cli.main.output_stream(stream, Mock())

    assert [(record.levelname, record.module, record.message) for record in caplog.records] == [
        ("error", "main", "Try 1/2: Could not open stream fake-stream (Could not open stream: failure)"),
        ("error", "main", "Try 2/2: Could not open stream fake-stream (Could not open stream: failure)"),
    ]
    assert not output.open.called, "Does not open the output on stream error"
    assert str(exc_info.value) == "Could not open stream fake-stream, tried 2 times, exiting"
    assert exc_info.value.code == 1


@pytest.mark.parametrize("argv", [["--retry-open=1"]], indirect=True)
@pytest.mark.parametrize("has_progress", [True, False])
def test_stream_runner_with_progress(
    monkeypatch: pytest.MonkeyPatch,
    caplog: pytest.LogCaptureFixture,
    argv: list,
    stream: Stream,
    output: PlayerOutput | FileOutput,
    has_progress: bool,
):
    streamio = BytesIO(b"0" * 8192 * 2)
    monkeypatch.setattr(stream, "open", Mock(return_value=streamio))

    mock_streamrunner = Mock()
    monkeypatch.setattr(streamlink_cli.main, "StreamRunner", mock_streamrunner)

    progress = sentinel.progress if has_progress else None
    monkeypatch.setattr(streamlink_cli.main, "get_output_progress", Mock(return_value=progress))

    assert streamlink_cli.main.output_stream(stream, Mock())

    assert [(record.levelname, record.module, record.message) for record in caplog.records] == [
        ("debug", "main", "Pre-buffering 8192 bytes"),
        ("debug", "main", "Writing stream to output"),
    ]
    assert mock_streamrunner.call_args_list == [call(streamio, output, progress=progress)]


filename = Path("filename")
file_output = FileOutput(filename)
file_recording = FileOutput(record=file_output)
player_output = PlayerOutput(Path("player"))
player_recording = PlayerOutput(Path("player"), record=file_output)


@pytest.mark.parametrize(
    ("argv", "supports_status_messages", "output", "expected"),
    [
        pytest.param(
            ["--progress=yes"],
            True,
            player_recording,
            {"path": filename},
            id="progress-status-messages-playeroutput-recording",
        ),
        pytest.param(
            ["--progress=yes"],
            True,
            file_output,
            {"path": filename},
            id="progress-status-messages-fileoutput",
        ),
        pytest.param(
            ["--progress=yes"],
            True,
            file_recording,
            {"path": filename},
            id="progress-status-messages-fileoutput-recording",
        ),
        pytest.param(
            ["--progress=yes"],
            True,
            player_output,
            {},
            id="progress-playeroutput-no-recording",
        ),
        pytest.param(
            ["--progress=yes"],
            True,
            file_output,
            {"path": filename},
            id="progress-fileoutput-no-recording",
        ),
        pytest.param(
            ["--progress=no"],
            True,
            file_output,
            {},
            id="no-progress-status-messages",
        ),
        pytest.param(
            ["--progress=yes"],
            False,
            file_output,
            {},
            id="progress-no-status-messages",
        ),
        pytest.param(
            ["--progress=no"],
            False,
            file_output,
            {},
            id="no-progress-no-status-messages",
        ),
        pytest.param(
            ["--progress=force"],
            False,
            file_output,
            {"path": filename, "interval": PROGRESS_INTERVAL_NO_STATUS, "status": False},
            id="force-progress-no-status-messages",
        ),
    ],
    indirect=["argv", "output"],
)
def test_get_output_progress(
    monkeypatch: pytest.MonkeyPatch,
    argv: list,
    output: PlayerOutput | FileOutput,
    supports_status_messages: bool,
    expected: dict,
):
    mock_progress = Mock(return_value=sentinel.progress if expected else None)
    monkeypatch.setattr(streamlink_cli.main, "Progress", mock_progress)

    mock_console = Mock(supports_status_messages=Mock(return_value=supports_status_messages))
    monkeypatch.setattr(streamlink_cli.main, "console", mock_console)

    result = streamlink_cli.main.get_output_progress(output)
    if not expected:
        assert result is None
    else:
        assert result is sentinel.progress
        assert mock_progress.call_args_list == [call(mock_console, **expected)]



"""
$description Live TV channels and video on-demand service from IndiHome TV, owned by Telkom Indonesia.
$url indihometv.com
$type live, vod
$region Indonesia
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(r"https?://(?:www\.)?indihometv\.com/"))
class IndiHomeTV(Plugin):
    def _get_streams(self):
        url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.any(
                    validate.all(
                        validate.xml_xpath_string("""
                            .//script[contains(text(), 'laylist.m3u8') or contains(text(), 'manifest.mpd')][1]/text()
                        """),
                        str,
                        re.compile(r"""(?P<q>['"])(?P<url>https://.*?/(?:[Pp]laylist\.m3u8|manifest\.mpd).+?)(?P=q)"""),
                        validate.none_or_all(
                            validate.get("url"),
                            validate.url(),
                        ),
                    ),
                    validate.all(
                        validate.xml_xpath_string(".//video[@id='video-player'][1]/source[1]/@src"),
                        validate.none_or_all(
                            validate.url(),
                        ),
                    ),
                ),
            ),
        )

        if url and ".m3u8" in url:
            return HLSStream.parse_variant_playlist(self.session, url)
        elif url and ".mpd" in url:
            return DASHStream.parse_manifest(self.session, url)


__plugin__ = IndiHomeTV



from __future__ import annotations

import subprocess
from collections.abc import Mapping, Sequence
from contextlib import AbstractContextManager, nullcontext
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest

from streamlink.exceptions import StreamlinkWarning
from streamlink_cli.output.player import (
    PlayerArgs,
    PlayerArgsMPV,
    PlayerArgsPotplayer,
    PlayerArgsVLC,
    PlayerOutput,
)


class TestPlayerArgs:
    @pytest.fixture()
    def playerargs(self, request: pytest.FixtureRequest):
        return PlayerOutput.playerargsfactory(**getattr(request, "param", {}))

    @pytest.fixture()
    def _playerargv(self, request: pytest.FixtureRequest, playerargs: PlayerArgs):
        assert playerargs.build() == getattr(request, "param", None)

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("playerargs", "_playerargv")
    @pytest.mark.parametrize(
        ("playerargs", "_playerargv"),
        [
            pytest.param(
                dict(path=Path("player")),
                ["player", "-"],
                id="Without player args",
            ),
            pytest.param(
                dict(path=Path("path to player"), args='p1 "1 2 3" p2="4 5 6" \'p3=7 8 9\' "a=\'b c\' d" \'e"f g"h\''),
                ["path to player", "p1", "1 2 3", "p2=4 5 6", "p3=7 8 9", "a='b c' d", 'e"f g"h', "-"],
                id="Player args tokenization",
            ),
            pytest.param(
                dict(path=Path("player"), args="param1 param2=value"),
                ["player", "param1", "param2=value", "-"],
                id="Implicit playerinput variable",
            ),
            pytest.param(
                dict(path=Path("player"), args="{playerinput} param1 param2=value"),
                ["player", "-", "param1", "param2=value"],
                id="Explicit playerinput variable",
            ),
            pytest.param(
                dict(path=Path("player"), args="{playerinput} param1 param2=value {playerinput}"),
                ["player", "-", "param1", "param2=value", "-"],
                id="Duplicate playerinput variable",
            ),
            pytest.param(
                dict(path=Path("player"), args="param1 {playerinput}-stdin param2"),
                ["player", "param1", "--stdin", "param2"],
                id="Combination of playerinput variable",
            ),
            pytest.param(
                dict(path=Path("player"), args="param1 param2=value {unknown}"),
                ["player", "param1", "param2=value", "{unknown}", "-"],
                id="Unknown player args variable",
            ),
            pytest.param(
                dict(path=Path("/usr/bin/player")),
                ["/usr/bin/player", "-"],
                marks=pytest.mark.posix_only,
                id="Absolute player path (POSIX)",
            ),
            pytest.param(
                dict(path=Path("C:\\path\\to\\player.exe")),
                ["C:\\path\\to\\player.exe", "-"],
                marks=pytest.mark.windows_only,
                id="Absolute player path (Windows)",
            ),
        ],
        indirect=True,
    )
    def test_argv(self):
        pass

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("playerargs", "_playerargv")
    @pytest.mark.parametrize(
        ("playerargs", "_playerargv"),
        [
            pytest.param(
                dict(path=Path("player")),
                ["player", "-"],
                id="stdin",
            ),
            pytest.param(
                dict(path=Path("player"), namedpipe=Mock(path="namedpipe")),
                ["player", "namedpipe"],
                id="namedpipe",
            ),
            pytest.param(
                dict(path=Path("player"), filename="https://localhost:65535/-._~:/?#[]@!$&'()*+,;="),
                ["player", "https://localhost:65535/-._~:/?#[]@!$&'()*+,;="],
                id="filename",
            ),
            pytest.param(
                dict(path=Path("player"), http=Mock(url="https://localhost:65535/")),
                ["player", "https://localhost:65535/"],
                id="http",
            ),
        ],
        indirect=True,
    )
    def test_input(self):
        pass

    @pytest.mark.parametrize(
        ("playerpath", "playerargsclass"),
        [
            pytest.param(
                "player",
                PlayerArgs,
                id="Unknown player",
            ),
            pytest.param(
                "vlc",
                PlayerArgsVLC,
                id="VLC",
            ),
            pytest.param(
                "vlc.exe",
                PlayerArgs,
                marks=pytest.mark.posix_only,
                id="VLC with .exe file extension (POSIX)",
            ),
            pytest.param(
                "vlc.exe",
                PlayerArgsVLC,
                marks=pytest.mark.windows_only,
                id="VLC with .exe file extension (Windows)",
            ),
            pytest.param(
                "/usr/bin/vlc",
                PlayerArgsVLC,
                marks=pytest.mark.posix_only,
                id="VLC with absolute path (POSIX)",
            ),
            pytest.param(
                "C:\\Program Files\\VideoLAN\\VLC\\vlc",
                PlayerArgsVLC,
                marks=pytest.mark.windows_only,
                id="VLC with absolute path (Windows)",
            ),
            pytest.param(
                "C:\\Program Files\\VideoLAN\\VLC\\vlc.exe",
                PlayerArgsVLC,
                marks=pytest.mark.windows_only,
                id="VLC with absolute path and file extension (Windows)",
            ),
            # don't repeat path and file extension tests
            # MPV
            pytest.param(
                "mpv",
                PlayerArgsMPV,
                id="MPV",
            ),
            # Potplayer
            pytest.param(
                "potplayer",
                PlayerArgsPotplayer,
                id="Potplayer (potplayer)",
            ),
            pytest.param(
                "potplayermini",
                PlayerArgsPotplayer,
                id="Potplayer (potplayermini)",
            ),
            pytest.param(
                "potplayermini64",
                PlayerArgsPotplayer,
                id="Potplayer (potplayermini64)",
            ),
        ],
    )
    def test_knownplayer(self, playerpath: str, playerargsclass: type[PlayerArgs]):
        assert isinstance(PlayerOutput.playerargsfactory(path=Path(playerpath)), playerargsclass)

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("playerargs", "_playerargv")
    @pytest.mark.parametrize(
        ("playerargs", "_playerargv"),
        [
            pytest.param(
                dict(path=Path("vlc"), namedpipe=Mock(path="namedpipe")),
                ["vlc", "namedpipe"],
                marks=pytest.mark.posix_only,
                id="VLC named pipe (POSIX)",
            ),
            pytest.param(
                dict(path=Path("vlc"), namedpipe=Mock(path="namedpipe")),
                ["vlc", "stream://\\namedpipe"],
                marks=pytest.mark.windows_only,
                id="VLC named pipe (Windows)",
            ),
            pytest.param(
                dict(path=Path("mpv"), namedpipe=Mock(path="namedpipe")),
                ["mpv", "namedpipe"],
                marks=pytest.mark.posix_only,
                id="MPV named pipe (POSIX)",
            ),
            pytest.param(
                dict(path=Path("mpv"), namedpipe=Mock(path="namedpipe")),
                ["mpv", "file://namedpipe"],
                marks=pytest.mark.windows_only,
                id="MPV named pipe (Windows)",
            ),
        ],
        indirect=True,
    )
    def test_knownplayer_input(self):
        pass

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("playerargs", "_playerargv")
    @pytest.mark.parametrize(
        ("playerargs", "_playerargv"),
        [
            pytest.param(
                dict(path=Path("player"), title="foo bar"),
                ["player", "-"],
                id="No title on unknown player",
            ),
            pytest.param(
                dict(path=Path("vlc"), title="foo bar: $a - $t"),
                ["vlc", "--input-title-format", "foo bar: $$a - $$t", "-"],
                id="VLC title",
            ),
            pytest.param(
                dict(path=Path("mpv"), title="foo bar"),
                ["mpv", "--force-media-title=foo bar", "-"],
                id="MPV title",
            ),
            pytest.param(
                dict(path=Path("potplayer"), title="foo bar"),
                ["potplayer", "-"],
                id="Potplayer title (stdin - no title)",
            ),
            pytest.param(
                dict(path=Path("potplayer"), title="foo bar", namedpipe=Mock(path="namedpipe")),
                ["potplayer", "namedpipe\\foo bar"],
                id="Potplayer title (namedpipe)",
            ),
            pytest.param(
                dict(path=Path("vlc"), args="param1 {playertitleargs} param2", title="foo bar"),
                ["vlc", "param1", "--input-title-format", "foo bar", "param2", "-"],
                id="Explicit playertitleargs variable",
            ),
            pytest.param(
                dict(path=Path("vlc"), args="param1{playertitleargs}param2", title="foo bar"),
                ["vlc", "param1--input-title-format", "foo barparam2", "-"],
                id="Explicit playertitleargs variable with improper usage (correct tokenization)",
            ),
        ],
        indirect=True,
    )
    def test_title(self):
        pass


# TODO: refactor PlayerOutput and write proper tests
class TestPlayerOutput:
    @pytest.fixture(autouse=True)
    def _os_environ(self, os_environ: dict[str, str]):
        os_environ["FAKE"] = "ENVIRONMENT"
        yield
        assert sorted(os_environ.keys()) == ["FAKE"], "Doesn't pollute the os.environ dict with custom env vars"

    @pytest.fixture()
    def playeroutput(self, request: pytest.FixtureRequest):
        with patch("streamlink_cli.output.player.sleep"):
            playeroutput = PlayerOutput(**getattr(request, "param", {}))
            yield playeroutput
            playeroutput.close()

    @pytest.fixture()
    def mock_which(self, request: pytest.FixtureRequest, playeroutput: PlayerOutput):
        with patch("streamlink_cli.output.player.which", return_value=getattr(request, "param", None)) as mock_which:
            yield mock_which

    @pytest.fixture()
    def mock_popen(self, playeroutput: PlayerOutput):
        with (
            patch("streamlink_cli.output.player.sleep"),
            patch("subprocess.Popen", return_value=Mock(poll=Mock(side_effect=Mock(return_value=None)))) as mock_popen,
        ):
            yield mock_popen

    @pytest.mark.parametrize(
        ("playeroutput", "mock_which", "args", "env", "logmessage"),
        [
            pytest.param(
                {"path": Path("player"), "args": "param1 param2", "call": False},
                "/resolved/player",
                ["/resolved/player", "param1", "param2", "-"],
                {"FAKE": "ENVIRONMENT"},
                "Opening subprocess: ['/resolved/player', 'param1', 'param2', '-']",
                id="Without custom env vars",
            ),
            pytest.param(
                {"path": Path("player"), "args": "param1 param2", "env": [("VAR1", "abc"), ("VAR2", "def")], "call": False},
                "/resolved/player",
                ["/resolved/player", "param1", "param2", "-"],
                {"FAKE": "ENVIRONMENT", "VAR1": "abc", "VAR2": "def"},
                "Opening subprocess: ['/resolved/player', 'param1', 'param2', '-'], env: {'VAR1': 'abc', 'VAR2': 'def'}",
                id="With custom env vars",
            ),
        ],
        indirect=["playeroutput", "mock_which"],
    )
    def test_open_popen_parameters(
        self,
        caplog: pytest.LogCaptureFixture,
        playeroutput: PlayerOutput,
        mock_which: Mock,
        mock_popen: Mock,
        args: Sequence[str],
        env: Mapping[str, str],
        logmessage: str,
    ):
        caplog.set_level(1, "streamlink")

        assert getattr(playeroutput, "player", None) is None

        playeroutput.open()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.cli.output", "debug", logmessage),
        ]
        assert mock_which.call_args_list == [call("player")]
        assert mock_popen.call_args_list == [
            call(
                args,
                bufsize=0,
                env=env,
                stdin=subprocess.PIPE,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            ),
        ]
        assert playeroutput.player is not None
        assert not playeroutput.player.terminate.called  # type: ignore

        playeroutput.close()
        assert playeroutput.player.terminate.call_count == 1  # type: ignore

    @pytest.mark.parametrize(
        ("playeroutput", "mock_which", "expected", "warns"),
        [
            pytest.param(
                dict(path=Path("foo")),
                "foo",
                nullcontext(),
                False,
                id="Player found",
            ),
            pytest.param(
                dict(path=Path("foo")),
                None,
                pytest.raises(FileNotFoundError, match=r"^Player executable not found$"),
                False,
                id="Player not found",
            ),
            pytest.param(
                dict(path=Path('"foo bar"')),
                None,
                pytest.raises(FileNotFoundError, match=r"^Player executable not found$"),
                True,
                id="Player not found with quotation warning",
            ),
        ],
        indirect=["playeroutput", "mock_which"],
    )
    def test_open_error(
        self,
        recwarn: pytest.WarningsRecorder,
        playeroutput: PlayerOutput,
        mock_which: Mock,
        mock_popen: Mock,
        expected: AbstractContextManager,
        warns: bool,
    ):
        with expected:
            playeroutput.open()
        assert any(record.category is StreamlinkWarning for record in recwarn.list) is warns



from streamlink.plugins.vkvideo import VKvideo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlVKvideo(PluginCanHandleUrl):
    __plugin__ = VKvideo

    should_match = [
        "https://live.vkvideo.ru/CHANNEL",
        "https://live.vkplay.ru/CHANNEL",
        "https://vkplay.live/CHANNEL",
    ]

    should_not_match = [
        "https://live.vkvideo.ru/",
        "https://live.vkplay.ru/",
        "https://vkplay.live/",
        "https://support.vkplay.ru/vkp_live",
        "https://vkplay.live/app/settings/edit",
        "https://vkplay.live/app/settings/external-apps",
    ]



import pytest
from requests_mock import Mocker

from streamlink import Streamlink
from streamlink.plugins.rtpplay import RTPPlay
from tests.plugins import PluginCanHandleUrl
from tests.resources import text


class TestPluginCanHandleUrlRTPPlay(PluginCanHandleUrl):
    __plugin__ = RTPPlay

    should_match = [
        "http://www.rtp.pt/play/",
        "https://www.rtp.pt/play/",
        "https://www.rtp.pt/play/direto/rtp1",
        "https://www.rtp.pt/play/direto/rtpmadeira",
    ]

    should_not_match = [
        "https://www.rtp.pt/programa/",
        "http://www.rtp.pt/programa/",
        "https://media.rtp.pt/",
        "http://media.rtp.pt/",
    ]


class TestRTPPlay:
    # all invalid HLS URLs at the beginning need to be ignored ("https://invalid")
    _content_pre = """
        /*  var player1 = new RTPPlayer({
                file: {hls : atob( decodeURIComponent(["aHR0c", "HM6Ly", "9pbnZ", "hbGlk"].join("") ) ), dash : foo() } }); */
        var f = {hls : atob( decodeURIComponent(["aHR0c", "HM6Ly", "9pbnZ", "hbGlk"].join("") ) ), dash: foo() };
    """
    # invalid resources sometimes have an empty string as HLS URL
    _content_invalid = """
        var f = {hls : ""};
    """
    # the actual HLS URL always comes last ("https://valid")
    _content_valid = """
        var f = {hls : decodeURIComponent(["https%3","A%2F%2F", "valid"].join("") ), dash: foo() };
    """
    _content_valid_b64 = """
        var f = {hls : atob( decodeURIComponent(["aHR0c", "HM6Ly", "92YWx", "pZA=="].join("") ) ), dash: foo() };
    """

    @pytest.fixture(scope="class")
    def playlist(self):
        with text("hls/test_master.m3u8") as fd:
            return fd.read()

    @pytest.fixture()
    def streams(self, request: pytest.FixtureRequest, requests_mock: Mocker, session: Streamlink, playlist: str):
        response = getattr(request, "param", "")

        requests_mock.get("https://www.rtp.pt/play/id/title", text=response)
        requests_mock.get("https://valid", text=playlist)
        requests_mock.get("https://invalid", exc=AssertionError)

        plugin = RTPPlay(session, "https://www.rtp.pt/play/id/title")

        return plugin._get_streams()

    @pytest.mark.parametrize(
        ("streams", "expected"),
        [
            pytest.param(
                "",
                False,
                id="empty",
            ),
            pytest.param(
                _content_pre + _content_invalid,
                False,
                id="invalid",
            ),
            pytest.param(
                _content_pre + _content_valid,
                True,
                id="valid",
            ),
            pytest.param(
                _content_pre + _content_valid_b64,
                True,
                id="valid-b64",
            ),
        ],
        indirect=["streams"],
    )
    def test_streams(self, streams, expected):
        assert (streams is not None) is expected



from __future__ import annotations

import ast
import re
from contextlib import nullcontext, suppress
from pathlib import Path
from textwrap import dedent
from typing import Any

import pytest

from build_backend.plugins_json import (
    PLUGINSJSON_COMMENTS,
    ParseConstantOrSequenceOfConstants,
    ParseMappingOfConstants,
    ParseSequenceOfConstants,
    PluginArgument,
    PluginMatcher,
    PluginVisitor,
    build,
    to_json,
)


does_not_raise = nullcontext()


class ConstantOrSequenceOfConstants(ParseConstantOrSequenceOfConstants):
    TYPE = bool, int, float, str
    REQUIRED = False


class SequenceOfConstants(ParseSequenceOfConstants):
    TYPE = bool, int, float, str
    REQUIRED = False


class MappingOfConstants(ParseMappingOfConstants):
    TYPE = bool, int, float, str
    REQUIRED = False


@pytest.mark.parametrize(
    ("parser", "code", "expected", "raises"),
    [
        pytest.param(
            SequenceOfConstants,
            """None""",
            None,
            does_not_raise,
            id="soc-optional",
        ),
        pytest.param(
            SequenceOfConstants,
            """[None,True,False,1,2,3.14,"a"]""",
            [None, True, False, 1, 2, 3.14, "a"],
            does_not_raise,
            id="soc-sequence",
        ),
        pytest.param(
            SequenceOfConstants,
            """1""",
            None,
            pytest.raises(ValueError, match=r"^Invalid constant type$"),
            id="soc-not-a-sequence",
        ),
        pytest.param(
            SequenceOfConstants,
            """[1,2,[]]""",
            None,
            pytest.raises(ValueError, match=r"^Invalid constant type$"),
            id="soc-no-nested-sequences",
        ),
        pytest.param(
            ConstantOrSequenceOfConstants,
            """None""",
            None,
            does_not_raise,
            id="cosoc-optional",
        ),
        pytest.param(
            ConstantOrSequenceOfConstants,
            """True""",
            True,
            does_not_raise,
            id="cosoc-constant",
        ),
        pytest.param(
            ConstantOrSequenceOfConstants,
            """[None,True,False,1,2,3.14,"a"]""",
            [None, True, False, 1, 2, 3.14, "a"],
            does_not_raise,
            id="cosoc-sequence",
        ),
        pytest.param(
            ConstantOrSequenceOfConstants,
            """[1,2,[]]""",
            None,
            pytest.raises(ValueError, match=r"^Invalid constant type$"),
            id="cosoc-no-nested-sequences",
        ),
        pytest.param(
            MappingOfConstants,
            """None""",
            None,
            does_not_raise,
            id="moc-optional",
        ),
        pytest.param(
            MappingOfConstants,
            """{"a": True, "b": 1, "c": 3.14, "d": "d", "e": [False, 2, 0.1, "x"]}""",
            {"a": True, "b": 1, "c": 3.14, "d": "d", "e": [False, 2, 0.1, "x"]},
            does_not_raise,
            id="moc-mapping",
        ),
        pytest.param(
            MappingOfConstants,
            """1""",
            None,
            pytest.raises(ValueError, match=r"^Invalid constant type$"),
            id="moc-not-a-mapping",
        ),
        pytest.param(
            MappingOfConstants,
            """[1]""",
            None,
            pytest.raises(ValueError, match=r"^Invalid constant type$"),
            id="moc-no-sequences",
        ),
        pytest.param(
            MappingOfConstants,
            """{"a": {}}""",
            None,
            pytest.raises(ValueError, match=r"^Invalid constant type$"),
            id="moc-no-nested-mappings",
        ),
    ],
)
def test_parse_sequence_or_mapping(parser: type[ast.NodeVisitor], code: str, expected: Any, raises: nullcontext):
    tree = ast.parse(code)
    item: ast.AST = tree.body[0].value  # type: ignore
    with raises:
        assert parser().visit(item) == expected


@pytest.mark.parametrize(
    ("code", "expected", "raises"),
    [
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"foo"))
                class NotAPlugin: pass
            """),
            [],
            does_not_raise,
            id="no-plugin-subclass",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"foo"))
                class NotAPlugin(GetBaseClass(), SomethingElse): pass
            """),
            [],
            does_not_raise,
            id="no-plugin-subclass2",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher
                class TestPlugin(Plugin): pass
            """),
            [],
            does_not_raise,
            id="no-decorator-call",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"foo"))
                @pluginmatcher(re.compile(pattern=r"foo"))
                @pluginmatcher(pattern=re.compile(r"foo"))
                @pluginmatcher(pattern=re.compile(pattern=r"foo"))
                @pluginmatcher(re.compile(r"bar", re.IGNORECASE))
                @pluginmatcher(re.compile(r"bar", flags=re.IGNORECASE))
                @pluginmatcher(pattern=re.compile(r"bar", re.IGNORECASE))
                @pluginmatcher(pattern=re.compile(r"bar", flags=re.IGNORECASE))
                @pluginmatcher(re.compile(r"baz", re.I | re.S | re.X))
                @pluginmatcher(re.compile(r"baz", flags=re.I | re.S | re.X))
                class TestPlugin(Plugin): pass
            """),
            [
                PluginMatcher(pattern="foo", flags=None, priority=None, name=None),
                PluginMatcher(pattern="foo", flags=None, priority=None, name=None),
                PluginMatcher(pattern="foo", flags=None, priority=None, name=None),
                PluginMatcher(pattern="foo", flags=None, priority=None, name=None),
                PluginMatcher(pattern="bar", flags=re.IGNORECASE, priority=None, name=None),
                PluginMatcher(pattern="bar", flags=re.IGNORECASE, priority=None, name=None),
                PluginMatcher(pattern="bar", flags=re.IGNORECASE, priority=None, name=None),
                PluginMatcher(pattern="bar", flags=re.IGNORECASE, priority=None, name=None),
                PluginMatcher(pattern="baz", flags=re.I | re.S | re.X, priority=None, name=None),
                PluginMatcher(pattern="baz", flags=re.I | re.S | re.X, priority=None, name=None),
            ],
            does_not_raise,
            id="pattern",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher()
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Missing @pluginmatcher pattern$"),
            id="pattern-missing",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(invalid)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher pattern: unknown AST node$"),
            id="pattern-invalid",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(get_pattern_object())
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher pattern: not a compiled regex$"),
            id="pattern-invalid-regex",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r'''  not a verbose regex pattern  '''))
                @pluginmatcher(re.compile(r'''
                    a\\sverbose[ ]+
                    regex\\ pattern
                ''', flags=re.VERBOSE))
                class TestPlugin(Plugin): pass
            """),
            [
                PluginMatcher(
                    pattern="  not a verbose regex pattern  ",
                    flags=None,
                    priority=None,
                    name=None,
                ),
                PluginMatcher(
                    pattern="a\\sverbose[ ]+regex\\ pattern",
                    flags=re.VERBOSE,
                    priority=None,
                    name=None,
                ),
            ],
            does_not_raise,
            id="pattern-pattern-verbose",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile())
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher pattern: missing pattern$"),
            id="pattern-pattern-missing",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(foo()))
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher pattern: unknown AST node$"),
            id="pattern-pattern-invalid-arg",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(pattern=foo()))
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher pattern: unknown AST node$"),
            id="pattern-pattern-invalid-kwarg",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"foo", 123))
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Unsupported node type$"),
            id="pattern-flag-invalid-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"foo", re.MULTILINE))
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid attribute$"),
            id="pattern-flag-invalid-attribute",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"foo", re.IGNORECASE & re.VERBOSE))
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Unsupported binary operator$"),
            id="pattern-flag-invalid-binary-operator",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"empty"))
                @pluginmatcher(re.compile(r"0"), 0)
                @pluginmatcher(re.compile(r"123"), priority=123)
                @pluginmatcher(re.compile(r"NO"), priority=NO_PRIORITY)
                @pluginmatcher(re.compile(r"LOW"), priority=LOW_PRIORITY)
                @pluginmatcher(re.compile(r"NORMAL"), priority=NORMAL_PRIORITY)
                @pluginmatcher(re.compile(r"HIGH"), priority=HIGH_PRIORITY)
                class TestPlugin(Plugin): pass
            """),
            [
                PluginMatcher(pattern="empty", flags=None, priority=None, name=None),
                PluginMatcher(pattern="0", flags=None, priority=0, name=None),
                PluginMatcher(pattern="123", flags=None, priority=123, name=None),
                PluginMatcher(pattern="NO", flags=None, priority=PluginMatcher.NO_PRIORITY, name=None),
                PluginMatcher(pattern="LOW", flags=None, priority=PluginMatcher.LOW_PRIORITY, name=None),
                PluginMatcher(pattern="NORMAL", flags=None, priority=PluginMatcher.NORMAL_PRIORITY, name=None),
                PluginMatcher(pattern="HIGH", flags=None, priority=PluginMatcher.HIGH_PRIORITY, name=None),
            ],
            does_not_raise,
            id="priority",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"unknown"), priority=foo())
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher priority: unknown AST node$"),
            id="priority-invalid-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"unknown"), priority=123.456)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher priority type$"),
            id="priority-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"unknown"), priority=UNKNOWN)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Unknown @pluginmatcher priority name$"),
            id="priority-unknown-name",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"one"), NORMAL_PRIORITY, "one")
                @pluginmatcher(re.compile(r"two"), NORMAL_PRIORITY, None)
                @pluginmatcher(re.compile(r"three"), name="three")
                @pluginmatcher(re.compile(r"four"), name=None)
                @pluginmatcher(name="five", pattern=re.compile(r"five"))
                class TestPlugin(Plugin): pass
            """),
            [
                PluginMatcher(pattern="one", flags=None, priority=PluginMatcher.NORMAL_PRIORITY, name="one"),
                PluginMatcher(pattern="two", flags=None, priority=PluginMatcher.NORMAL_PRIORITY, name=None),
                PluginMatcher(pattern="three", flags=None, priority=None, name="three"),
                PluginMatcher(pattern="four", flags=None, priority=None, name=None),
                PluginMatcher(pattern="five", flags=None, priority=None, name="five"),
            ],
            does_not_raise,
            id="name",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"."), name=NAME)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher name: unknown AST node$"),
            id="name-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"."), name=123)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginmatcher name type$"),
            id="name-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"."), NORMAL_PRIORITY, "name", 123)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid number of arguments$"),
            id="invalid-arguments-number",
        ),
        pytest.param(
            dedent("""
                @pluginmatcher(re.compile(r"."), invalid=123)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid keyword argument$"),
            id="invalid-keyword",
        ),
    ],
)
def test_pluginmatcher(code: str, expected: list, raises: nullcontext):
    tree = ast.parse(code)
    pluginvisitor = PluginVisitor()
    with raises:
        pluginvisitor.visit(tree)
        assert pluginvisitor.matchers == expected


@pytest.mark.parametrize(
    ("code", "expected", "raises"),
    [
        pytest.param(
            dedent("""
                @pluginargument("foo")
                @pluginargument(name="bar")
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo"),
                PluginArgument(name="bar"),
            ],
            does_not_raise,
            id="name",
        ),
        pytest.param(
            dedent("""
                @pluginargument()
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Missing @pluginargument name$"),
            id="name-missing",
        ),
        pytest.param(
            dedent("""
                @pluginargument(name=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument name: unknown AST node$"),
            id="name-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument(name=123)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument name type$"),
            id="name-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", action=None)
                @pluginargument("bar", action="store_true")
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", action=None),
                PluginArgument(name="bar", action="store_true"),
            ],
            does_not_raise,
            id="action",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", action=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument action: unknown AST node$"),
            id="action-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", action=123)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument action type$"),
            id="action-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", nargs=None)
                @pluginargument("bar", nargs=2)
                @pluginargument("baz", nargs="?")
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", nargs=None),
                PluginArgument(name="bar", nargs=2),
                PluginArgument(name="baz", nargs="?"),
            ],
            does_not_raise,
            id="nargs",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", nargs=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument nargs: unknown AST node$"),
            id="nargs-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", nargs=123.456)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument nargs type$"),
            id="nargs-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", const=None)
                @pluginargument("bar", const=1)
                @pluginargument("baz", const=["a", 2, 3.14, True])
                @pluginargument("qux", const=("a", 2, 3.14, True))
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", const=None),
                PluginArgument(name="bar", const=1),
                PluginArgument(name="baz", const=["a", 2, 3.14, True]),
                PluginArgument(name="qux", const=["a", 2, 3.14, True]),
            ],
            does_not_raise,
            id="const",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", const=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument const: unknown AST node$"),
            id="const-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", const=[[]])
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument const type$"),
            id="const-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", default=None)
                @pluginargument("bar", default=1)
                @pluginargument("baz", default=["a", 2, 3.14, True])
                @pluginargument("qux", default=("a", 2, 3.14, True))
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", default=None),
                PluginArgument(name="bar", default=1),
                PluginArgument(name="baz", default=["a", 2, 3.14, True]),
                PluginArgument(name="qux", default=["a", 2, 3.14, True]),
            ],
            does_not_raise,
            id="default",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", default=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument default: unknown AST node$"),
            id="default-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", default=[[]])
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument default type$"),
            id="default-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type=None)
                @pluginargument("bar", type="comma_list")
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", type=None),
                PluginArgument(name="bar", type="comma_list"),
            ],
            does_not_raise,
            id="type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument type: unknown AST node$"),
            id="type-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type=123)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument type type$"),
            id="type-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type_args=None)
                @pluginargument("bar", type_args=["a", 2, 3.14, True])
                @pluginargument("baz", type_args=("a", 2, 3.14, True))
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", type_args=None),
                PluginArgument(name="bar", type_args=["a", 2, 3.14, True]),
                PluginArgument(name="baz", type_args=["a", 2, 3.14, True]),
            ],
            does_not_raise,
            id="type-args",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type_args=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument type_args: unknown AST node$"),
            id="type-args-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type_args=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument type_args type$"),
            id="type-args-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type_kwargs=None)
                @pluginargument("bar", type_kwargs={"a": True, "b": 1, "c": 3.14, "d": "d", "e": [False, 2, 0.1, "x"]})
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", type_kwargs=None),
                PluginArgument(name="bar", type_kwargs={"a": True, "b": 1, "c": 3.14, "d": "d", "e": [False, 2, 0.1, "x"]}),
            ],
            does_not_raise,
            id="type-kwargs",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type_kwargs=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument type_kwargs: unknown AST node$"),
            id="type-kwargs-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", type_kwargs=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument type_kwargs type$"),
            id="type-kwargs-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", choices=None)
                @pluginargument("bar", choices=["a", 2, 3.14, True])
                @pluginargument("baz", choices=("a", 2, 3.14, True))
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", choices=None),
                PluginArgument(name="bar", choices=["a", 2, 3.14, True]),
                PluginArgument(name="baz", choices=["a", 2, 3.14, True]),
            ],
            does_not_raise,
            id="choices",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", choices=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument choices: unknown AST node$"),
            id="choices-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", choices=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument choices type$"),
            id="choices-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", required=None)
                @pluginargument("bar", required=True)
                @pluginargument("baz", required=False)
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", required=None),
                PluginArgument(name="bar", required=True),
                PluginArgument(name="baz", required=False),
            ],
            does_not_raise,
            id="required",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", required=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument required: unknown AST node$"),
            id="required-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", required=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument required type$"),
            id="required-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", help=argparse.SUPPRESS)
                @pluginargument("bar", help="==SUPPRESS==")
                @pluginargument("baz", help="help")
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", help="==SUPPRESS=="),
                PluginArgument(name="bar", help="==SUPPRESS=="),
                PluginArgument(name="baz", help="help"),
            ],
            does_not_raise,
            id="help",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", help=other.SUPPRESS)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument help type$"),
            id="help-not-argparse-suppress",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", help=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument help: unknown AST node$"),
            id="help-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", help=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument help type$"),
            id="help-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", metavar=None)
                @pluginargument("bar", metavar="bar")
                @pluginargument("baz", metavar=["foo", "bar"])
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", metavar=None),
                PluginArgument(name="bar", metavar="bar"),
                PluginArgument(name="baz", metavar=["foo", "bar"]),
            ],
            does_not_raise,
            id="metavar",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", metavar=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument metavar: unknown AST node$"),
            id="metavar-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", metavar=[1])
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument metavar type$"),
            id="metavar-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", dest=None)
                @pluginargument("bar", dest="bar")
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", dest=None),
                PluginArgument(name="bar", dest="bar"),
            ],
            does_not_raise,
            id="dest",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", dest=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument dest: unknown AST node$"),
            id="dest-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", dest=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument dest type$"),
            id="dest-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", requires=None)
                @pluginargument("bar", requires="foo")
                @pluginargument("baz", requires=["foo", "bar"])
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", requires=None),
                PluginArgument(name="bar", requires="foo"),
                PluginArgument(name="baz", requires=["foo", "bar"]),
            ],
            does_not_raise,
            id="requires",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", requires=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument requires: unknown AST node$"),
            id="requires-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", requires=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument requires type$"),
            id="requires-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", prompt=None)
                @pluginargument("bar", prompt="bar")
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", prompt=None),
                PluginArgument(name="bar", prompt="bar"),
            ],
            does_not_raise,
            id="prompt",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", prompt=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument prompt: unknown AST node$"),
            id="prompt-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", prompt=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument prompt type$"),
            id="prompt-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", sensitive=None)
                @pluginargument("bar", sensitive=True)
                @pluginargument("baz", sensitive=False)
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", sensitive=None),
                PluginArgument(name="bar", sensitive=True),
                PluginArgument(name="baz", sensitive=False),
            ],
            does_not_raise,
            id="sensitive",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", sensitive=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument sensitive: unknown AST node$"),
            id="sensitive-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", sensitive=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument sensitive type$"),
            id="sensitive-invalid-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", argument_name=None)
                @pluginargument("bar", argument_name="bar")
                class TestPlugin(Plugin): pass
            """),
            [
                PluginArgument(name="foo", argument_name=None),
                PluginArgument(name="bar", argument_name="bar"),
            ],
            does_not_raise,
            id="argument-name",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", argument_name=INVALID)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument argument_name: unknown AST node$"),
            id="argument-name-unknown-node-type",
        ),
        pytest.param(
            dedent("""
                @pluginargument("foo", argument_name=1)
                class TestPlugin(Plugin): pass
            """),
            None,
            pytest.raises(ValueError, match=r"^Invalid @pluginargument argument_name type$"),
            id="argument-name-invalid-type",
        ),
    ],
)
def test_pluginargument(code: str, expected: list, raises: nullcontext):
    tree = ast.parse(code)
    pluginvisitor = PluginVisitor()
    with raises:
        pluginvisitor.visit(tree)
        assert pluginvisitor.arguments == expected


@pytest.mark.parametrize(
    ("code", "expected"),
    [
        pytest.param(
            dedent("""
                class TestPlugin(Plugin): pass
            """),
            False,
            id="does-not-export",
        ),
        pytest.param(
            dedent("""
                class TestPlugin(Plugin): pass
                __plugin__ = TestPlugin
            """),
            True,
            id="exports",
        ),
    ],
)
def test_exports(code: str, expected: bool):
    tree = ast.parse(code)
    pluginvisitor = PluginVisitor()
    pluginvisitor.visit(tree)
    assert pluginvisitor.exports == expected


@pytest.fixture()
def test_plugin_code(request: pytest.FixtureRequest):
    faulty = getattr(request, "param", {}).get("faulty", False)

    return dedent(f"""
        import argparse
        import re

        from streamlink.plugin import LOW_PRIORITY, Plugin, pluginargument, pluginmatcher


        @pluginmatcher(re.compile("https://a"))
        @pluginmatcher(name="b", pattern=re.compile("https://b", re.IGNORECASE), priority=LOW_PRIORITY)
        @pluginmatcher(name="c", pattern=re.compile("https://c", re.I | re.X), priority=20)
        @pluginargument(
            "a",
            type="comma_list_filter",
            type_kwargs={{"acceptable": ["a", "b"]}},
            default=["a"],
            help="a",
        )
        @pluginargument(
            "b",
            requires="a",
            sensitive={"True" if not faulty else "INVALID"},
            help=argparse.SUPPRESS,
        )
        class TestPlugin(Plugin):
            def _get_streams(self):
                return None

        __plugin__ = TestPlugin
    """).strip()


@pytest.fixture()
def test_plugins_dir(tmp_path: Path, test_plugin_code: str):
    file_path = tmp_path / "testplugin.py"
    try:
        with file_path.open("w", encoding="utf-8") as fp:
            fp.write(test_plugin_code)
        yield tmp_path
    finally:
        with suppress(FileNotFoundError):
            file_path.unlink()


def test_plugin(test_plugin_code: str):
    tree = ast.parse(test_plugin_code)
    pluginvisitor = PluginVisitor()
    pluginvisitor.visit(tree)

    assert pluginvisitor.exports
    assert pluginvisitor.matchers == [
        PluginMatcher(pattern="https://a", flags=None, priority=None, name=None),
        PluginMatcher(pattern="https://b", flags=re.IGNORECASE, priority=PluginMatcher.LOW_PRIORITY, name="b"),
        PluginMatcher(pattern="https://c", flags=re.I | re.X, priority=20, name="c"),
    ]
    assert pluginvisitor.arguments == [
        PluginArgument(name="a", type="comma_list_filter", type_kwargs={"acceptable": ["a", "b"]}, default=["a"], help="a"),
        PluginArgument(name="b", requires="a", sensitive=True, help="==SUPPRESS=="),
    ]


@pytest.mark.parametrize(
    ("comments", "expected_comments"),
    [
        pytest.param(
            None,
            PLUGINSJSON_COMMENTS,
            id="default",
        ),
        pytest.param(
            ["foo", "bar"],
            ["foo", "bar"],
            id="custom",
        ),
        pytest.param(
            [],
            [],
            id="empty",
        ),
    ],
)
def test_build(
    capsys: pytest.CaptureFixture,
    test_plugins_dir: Path,
    comments: list[str] | None,
    expected_comments: list[str],
):
    data = build(test_plugins_dir)
    to_json(data, comments=comments, pretty=True)
    out, _err = capsys.readouterr()
    assert (
        out
        == "".join(f"// {comment}\n" for comment in expected_comments)
        + dedent(
            """
                {
                  "testplugin": {
                    "matchers": [
                      {
                        "pattern": "https://a"
                      },
                      {
                        "pattern": "https://b",
                        "flags": 2,
                        "priority": 10,
                        "name": "b"
                      },
                      {
                        "pattern": "https://c",
                        "flags": 66,
                        "priority": 20,
                        "name": "c"
                      }
                    ],
                    "arguments": [
                      {
                        "name": "a",
                        "default": [
                          "a"
                        ],
                        "type": "comma_list_filter",
                        "type_kwargs": {
                          "acceptable": [
                            "a",
                            "b"
                          ]
                        },
                        "help": "a"
                      },
                      {
                        "name": "b",
                        "help": "==SUPPRESS==",
                        "requires": "a",
                        "sensitive": true
                      }
                    ]
                  }
                }
            """,
        ).strip()
    )


@pytest.mark.parametrize("test_plugin_code", [{"faulty": True}], indirect=True)
def test_build_faulty(capsys: pytest.CaptureFixture, test_plugins_dir: Path, test_plugin_code: str):
    with pytest.raises(SyntaxError) as cm:
        build(test_plugins_dir)
    out, err = capsys.readouterr()
    assert not out
    assert not err

    assert cm.value
    msg, (filename, lineno, column, line) = cm.value.args
    assert msg == "Invalid @pluginargument sensitive: unknown AST node"
    assert filename == "testplugin.py"
    assert lineno == 20
    assert column == 15
    assert line == "    sensitive=INVALID,"
    assert cm.value.__cause__ is None



from streamlink.plugins.sportschau import Sportschau
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSportschau(PluginCanHandleUrl):
    __plugin__ = Sportschau

    should_match = [
        "https://www.sportschau.de/fussball/uefa-euro-2024/spaniens-europameister-begeistert-empfangen,em-spanien-feier-100.html",
        "https://www.sportschau.de/olympia/live/schwimmen-finals-m-f,livestream-olympia-schwimmen-110.html",
        "https://www.sportschau.de/podcasts/sportschau-olympia-podcast/tag-6-vier-gewinnt,audio-tag-6-vier-gewinnt-100.html",
    ]



from streamlink.plugins.douyin import Douyin
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlDouyin(PluginCanHandleUrl):
    __plugin__ = Douyin

    should_match_groups = [
        ("https://live.douyin.com/ROOM_ID", {"room_id": "ROOM_ID"}),
    ]

    should_not_match = [
        "https://live.douyin.com",
    ]



from streamlink.plugins.aloula import Aloula
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlAloula(PluginCanHandleUrl):
    __plugin__ = Aloula

    should_match_groups = [
        (("live", "https://aloula.sba.sa/live/saudiatv"), {"live_slug": "saudiatv"}),
        (("live", "https://www.aloula.sa/live/saudiatv"), {"live_slug": "saudiatv"}),
        (("live", "https://aloula.sba.sa/en/live/saudiatv"), {"live_slug": "saudiatv"}),
        (("live", "https://www.aloula.sa/en/live/saudiatv"), {"live_slug": "saudiatv"}),
        (("vod", "https://aloula.sba.sa/episode/6676"), {"vod_id": "6676"}),
        (("vod", "https://www.aloula.sa/episode/6676"), {"vod_id": "6676"}),
        (("vod", "https://aloula.sba.sa/en/episode/6676"), {"vod_id": "6676"}),
        (("vod", "https://www.aloula.sa/en/episode/6676"), {"vod_id": "6676"}),
    ]

    should_not_match = [
        "https://www.aloula.sa/en/any",
        "https://www.aloula.sa/de/any/path",
        "https://www.aloula.sa/live/",
        "https://www.aloula.sa/abc/live/slug",
        "https://www.aloula.sa/en/live/",
        "https://www.aloula.sa/episode/",
        "https://www.aloula.sa/abc/episode/123",
        "https://www.aloula.sa/en/episode/",
        "https://www.aloula.sa/episode/abc",
        "https://www.aloula.sa/de/episode/abc",
    ]
